#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.addons import openscad
from ezdxf.render import MeshBuilder
from ezdxf.math import Matrix44


def test_matrix44_to_multmatrix_str():
    s = openscad.str_matrix44(Matrix44.translate(20, 30, 40))
    assert (
        s == "[[1.0, 0.0, 0.0, 20.0],"
        " [0.0, 1.0, 0.0, 30.0],"
        " [0.0, 0.0, 1.0, 40.0],"
        " [0.0, 0.0, 0.0, 1.0]]"
    )


class TestScriptBuilder:
    def test_build_boolean_operation(self):
        m = MeshBuilder()
        script = openscad.boolean_operation(openscad.UNION, m, m)
        assert script == (
            "union() {\n"
            "polyhedron(points = [\n"
            "], faces = [\n"
            "], convexity = 10);\n"
            "\n"
            "polyhedron(points = [\n"
            "], faces = [\n"
            "], convexity = 10);\n"
            "\n"
            "}"
        )

    @pytest.mark.parametrize(
        "op,func",
        [
            (openscad.UNION, "union()"),
            (openscad.DIFFERENCE, "difference()"),
            (openscad.INTERSECTION, "intersection()"),
        ],
    )
    def test_build_all_operations(self, op, func):
        m = MeshBuilder()
        assert openscad.boolean_operation(op, m, m).startswith(func)

    def test_add_multmatrix_operation(self):
        script = openscad.Script()
        script.add_multmatrix(Matrix44.translate(20, 30, 40))
        result = script.get_string()
        assert result.startswith("multmatrix(m = [[")
        assert result.endswith("]])"), "operation without a pending ';'"

    def test_add_translation_operation(self):
        script = openscad.Script()
        script.add_translate((10, 20, 30))
        assert script.get_string() == "translate(v = [10, 20, 30])"

    def test_add_rotation_operation(self):
        script = openscad.Script()
        script.add_rotate(10, 20, 30)
        assert script.get_string() == "rotate(a = [10, 20, 30])"

    def test_add_rotation_about_axis_operation(self):
        script = openscad.Script()
        script.add_rotate_about_axis(40, (10, 20, 30))
        assert script.get_string() == "rotate(a = 40, v = [10, 20, 30])"

    def test_add_scale_operation(self):
        script = openscad.Script()
        script.add_scale(10, 20, 30)
        assert script.get_string() == "scale(v = [10, 20, 30])"

    def test_add_resize_operation(self):
        script = openscad.Script()
        script.add_resize(10, 20, 30)
        assert script.get_string() == "resize(newsize = [10, 20, 30])"

    def test_add_auto_resize_operation(self):
        script = openscad.Script()
        script.add_resize(10, 0, 0, auto=True)
        assert (
            script.get_string() == "resize(newsize = [10, 0, 0], auto = true)"
        )

    def test_add_multi_auto_resize_operation(self):
        script = openscad.Script()
        script.add_resize(10, 0, 0, auto=(False, True, False))
        assert (
            script.get_string()
            == "resize(newsize = [10, 0, 0], auto = [false, true, false])"
        )

    def test_add_mirror_operation(self):
        script = openscad.Script()
        script.add_mirror((10, 0, 0))
        assert script.get_string() == "mirror(v = [1, 0, 0])"


def test_str_polygon():
    s = openscad.str_polygon([(0, 0), (0, 0)])
    assert s == "polygon(points = [\n  [0, 0],\n  [0, 0],\n],\nconvexity = 10);"


def test_str_polygon_is_automatically_closed():
    s = openscad.str_polygon([(0, 0), (1, 0)])
    assert s == "polygon(points = [\n  [0, 0],\n  [1, 0],\n  [0, 0],\n],\nconvexity = 10);"


def test_str_polygon_with_holes():
    s = openscad.str_polygon(
        [(0, 0), (0, 0)], holes=([[1, 1], [1, 1]], [[2, 2], [2, 2]])
    )
    assert (
        s == "polygon(points = [\n"  # list of 6 vertices
        "  [0, 0],\n"
        "  [0, 0],\n"
        "  [1, 1],\n"
        "  [1, 1],\n"
        "  [2, 2],\n"
        "  [2, 2],\n"
        "],\n"
        "paths = [\n"  # list of 3 paths
        "  [0, 1],\n"  # list if vertex indices of exterior path 
        "  [2, 3],\n"  # list of vertex indices of the 1st hole 
        "  [4, 5],\n"  # list of vertex indices of the 2nd hole
        "],\n"
        "convexity = 10);"
    )


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.layouts import VirtualLayout
from ezdxf.enums import TextEntityAlignment


@pytest.fixture
def msp():
    return VirtualLayout()


def test_default_settings(msp):
    line = msp.add_line((0, 0), (1, 1))
    assert line.dxf.layer == "0"
    assert line.dxf.color == 256
    assert line.dxf.linetype == "BYLAYER"
    assert line.dxf.ltscale == 1.0
    assert line.dxf.invisible == 0
    assert line.dxf.extrusion == (0, 0, 1)
    line.dxf.lineweight = 18  # set line weight
    assert line.dxf.lineweight == 18  # get line weight


@pytest.fixture
def line(msp):
    return msp.add_line((0, 0), (1, 1))


def test_ac1018_default_settings(line):
    line = line
    assert line.has_dxf_attrib("true_color") is False  # no default true color
    assert line.has_dxf_attrib("color_name") is False  # no default color name
    assert (
        line.has_dxf_attrib("transparency") is False
    )  # no default transparency


def test_ac1018_true_color(line):
    line.dxf.true_color = 0x0F0F0F
    assert 0x0F0F0F == line.dxf.true_color
    assert (
        0x0F,
        0x0F,
        0x0F,
    ) == line.rgb  # shortcut for modern graphic entities
    line.rgb = (255, 255, 255)  # shortcut for modern graphic entities
    assert 0xFFFFFF == line.dxf.true_color


def test_ac1018_color_name(line):
    line.dxf.color_name = "Rot"
    assert "Rot" == line.dxf.color_name


def test_ac1018_transparency(line):
    line.dxf.transparency = 0x020000FF  # 0xFF = opaque; 0x00 = 100% transparent
    assert 0x020000FF == line.dxf.transparency
    # recommend usage: helper property ModernGraphicEntity.transparency
    assert 0.0 == line.transparency  # 0. =  opaque; 1. = 100% transparent
    line.transparency = 1.0
    assert 0x02000000 == line.dxf.transparency


def test_ac1021_default_settings(line):
    assert line.has_dxf_attrib("shadow_mode") is False  # no default shadow_mode


@pytest.fixture(scope="module")
def layout():
    # Do not test VirtualLayout() here
    import ezdxf

    doc = ezdxf.new()
    return doc.modelspace()


def test_create_line(layout):
    line = layout.add_line((0, 0), (1, 1))
    assert (0.0, 0.0) == line.dxf.start
    assert (1.0, 1.0) == line.dxf.end


def test_create_circle(layout):
    circle = layout.add_circle((3, 3), 5)
    assert (3.0, 3.0) == circle.dxf.center
    assert 5.0 == circle.dxf.radius


def test_create_arc(layout):
    arc = layout.add_arc((3, 3), 5, 30, 60)
    assert (3.0, 3.0) == arc.dxf.center
    assert 5.0 == arc.dxf.radius
    assert 30.0 == arc.dxf.start_angle
    assert 60.0 == arc.dxf.end_angle


def test_create_trace(layout):
    trace = layout.add_trace([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert (0, 0) == trace[0]
    assert (1, 0) == trace.dxf.vtx1
    assert (1, 1) == trace[2]
    assert (0, 1) == trace.dxf.vtx3


def test_create_solid(layout):
    trace = layout.add_solid([(0, 0), (1, 0), (1, 1)])
    assert (0, 0) == trace.dxf.vtx0
    assert (1, 0) == trace[1]
    assert (1, 1) == trace.dxf.vtx2
    assert (1, 1) == trace[3]


def test_create_3dface(layout):
    trace = layout.add_3dface([(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)])
    assert (0, 0, 0) == trace.dxf.vtx0
    assert (1, 0, 0) == trace[1]
    assert (1, 1, 0) == trace.dxf.vtx2
    assert (0, 1, 0) == trace[3]


def test_create_text(layout):
    text = layout.add_text("text")
    assert "text" == text.dxf.text


def test_text_set_alignment(layout):
    text = layout.add_text("text")
    text.set_placement((2, 2), align=TextEntityAlignment.TOP_CENTER)
    assert 1 == text.dxf.halign
    assert 3 == text.dxf.valign
    assert (2, 2) == text.dxf.align_point


def test_text_set_fit_alignment(layout):
    text = layout.add_text("text")
    text.set_placement((2, 2), (4, 2), align=TextEntityAlignment.FIT)
    assert 5 == text.dxf.halign
    assert 0 == text.dxf.valign
    assert (2, 2) == text.dxf.insert
    assert (4, 2) == text.dxf.align_point


def test_text_get_alignment(layout):
    text = layout.add_text("text")
    text.dxf.halign = 1
    text.dxf.valign = 3
    assert text.get_align_enum() == TextEntityAlignment.TOP_CENTER


def test_text_get_pos_TOP_CENTER(layout):
    text = layout.add_text("text")
    text.set_placement((2, 2), align=TextEntityAlignment.TOP_CENTER)
    align, p1, p2 = text.get_placement()
    assert align == TextEntityAlignment.TOP_CENTER
    assert p1 == (2, 2)
    assert p2 is None


def test_text_get_pos_LEFT(layout):
    text = layout.add_text("text")
    text.set_placement((2, 2))
    align, p1, p2 = text.get_placement()
    assert align == TextEntityAlignment.LEFT
    assert p1 == (2, 2)
    assert p2 is None


def test_create_shape(layout):
    shape = layout.add_shape("TestShape", (1, 2), 3.0)
    assert "TestShape" == shape.dxf.name
    assert (1.0, 2.0) == shape.dxf.insert
    assert 3.0 == shape.dxf.size
    assert 0.0 == shape.dxf.rotation
    assert 1.0 == shape.dxf.xscale
    assert 0.0 == shape.dxf.oblique


def test_create_ray(layout):
    ray = layout.add_ray((1, 2, 0), (1, 0, 0))
    assert (1, 2, 0) == ray.dxf.start
    assert (1, 0, 0) == ray.dxf.unit_vector


def test_xline_ray(layout):
    xline = layout.add_xline((1, 2, 0), (1, 0, 0))
    assert (1, 2, 0) == xline.dxf.start
    assert (1, 0, 0) == xline.dxf.unit_vector



# Copyright (c) 2011-2024, Manfred Moitzi
# License: MIT License
# These are the pure Python implementations of the Cython accelerated
# construction tools: ezdxf/acc/construct.pyx
from __future__ import annotations
from typing import Iterable, Sequence, Optional, TYPE_CHECKING
import math

# The pure Python implementation can't import from ._ctypes or ezdxf.math!
from ._vector import Vec2, Vec3

if TYPE_CHECKING:
    from ezdxf.math import UVec

TOLERANCE = 1e-10
RAD_ABS_TOL = 1e-15
DEG_ABS_TOL = 1e-13


def has_clockwise_orientation(vertices: Iterable[UVec]) -> bool:
    """Returns ``True`` if the given 2D `vertices` have clockwise orientation.
    Ignores the z-axis of all vertices.

    Args:
        vertices: iterable of :class:`Vec2` compatible objects

    Raises:
        ValueError: less than 3 vertices

    """
    vertices = Vec2.list(vertices)
    if len(vertices) < 3:
        raise ValueError("At least 3 vertices required.")

    # close polygon:
    if not vertices[0].isclose(vertices[-1]):
        vertices.append(vertices[0])

    return (
        sum(
            (p2.x - p1.x) * (p2.y + p1.y)
            for p1, p2 in zip(vertices, vertices[1:])
        )
        > 0.0
    )


def intersection_line_line_2d(
    line1: Sequence[Vec2],
    line2: Sequence[Vec2],
    virtual=True,
    abs_tol=TOLERANCE,
) -> Optional[Vec2]:
    """
    Compute the intersection of two lines in the xy-plane.

    Args:
        line1: start- and end point of first line to test
            e.g. ((x1, y1), (x2, y2)).
        line2: start- and end point of second line to test
            e.g. ((x3, y3), (x4, y4)).
        virtual: ``True`` returns any intersection point, ``False`` returns
            only real intersection points.
        abs_tol: tolerance for intersection test.

    Returns:
        ``None`` if there is no intersection point (parallel lines) or
        intersection point as :class:`Vec2`

    """
    # Algorithm based on: http://paulbourke.net/geometry/pointlineplane/
    # chapter: Intersection point of two line segments in 2 dimensions
    s1, s2 = line1  # the subject line
    c1, c2 = line2  # the clipping line
    s1x = s1.x
    s1y = s1.y
    s2x = s2.x
    s2y = s2.y
    c1x = c1.x
    c1y = c1.y
    c2x = c2.x
    c2y = c2.y

    den = (c2y - c1y) * (s2x - s1x) - (c2x - c1x) * (s2y - s1y)
    if math.fabs(den) <= abs_tol:
        return None

    us = ((c2x - c1x) * (s1y - c1y) - (c2y - c1y) * (s1x - c1x)) / den
    intersection_point = Vec2(s1x + us * (s2x - s1x), s1y + us * (s2y - s1y))
    if virtual:
        return intersection_point

    # 0 = intersection point is the start point of the line
    # 1 = intersection point is the end point of the line
    # otherwise: linear interpolation
    lwr = 0.0  # tolerances required?
    upr = 1.0  # tolerances required?
    if lwr <= us <= upr:  # intersection point is on the subject line
        uc = ((s2x - s1x) * (s1y - c1y) - (s2y - s1y) * (s1x - c1x)) / den
        if lwr <= uc <= upr:  # intersection point is on the clipping line
            return intersection_point
    return None


def _determinant(v1, v2, v3) -> float:
    """Returns determinant."""
    e11, e12, e13 = v1
    e21, e22, e23 = v2
    e31, e32, e33 = v3

    return (
        e11 * e22 * e33
        + e12 * e23 * e31
        + e13 * e21 * e32
        - e13 * e22 * e31
        - e11 * e23 * e32
        - e12 * e21 * e33
    )


def intersection_ray_ray_3d(
    ray1: Sequence[Vec3], ray2: Sequence[Vec3], abs_tol=TOLERANCE
) -> Sequence[Vec3]:
    """
    Calculate intersection of two 3D rays, returns a 0-tuple for parallel rays,
    a 1-tuple for intersecting rays and a 2-tuple for not intersecting and not
    parallel rays with points of the closest approach on each ray.

    Args:
        ray1: first ray as tuple of two points as :class:`Vec3` objects
        ray2: second ray as tuple of two points as :class:`Vec3` objects
        abs_tol: absolute tolerance for comparisons

    """
    # source: http://www.realtimerendering.com/intersections.html#I304
    o1, p1 = ray1
    d1 = (p1 - o1).normalize()
    o2, p2 = ray2
    d2 = (p2 - o2).normalize()
    d1xd2 = d1.cross(d2)
    denominator = d1xd2.magnitude_square
    if denominator <= abs_tol:
        # ray1 is parallel to ray2
        return tuple()
    else:
        o2_o1 = o2 - o1
        det1 = _determinant(o2_o1, d2, d1xd2)
        det2 = _determinant(o2_o1, d1, d1xd2)
        p1 = o1 + d1 * (det1 / denominator)
        p2 = o2 + d2 * (det2 / denominator)
        if p1.isclose(p2, abs_tol=abs_tol):
            # ray1 and ray2 have an intersection point
            return (p1,)
        else:
            # ray1 and ray2 do not have an intersection point,
            # p1 and p2 are the points of closest approach on each ray
            return p1, p2


def arc_angle_span_deg(start: float, end: float) -> float:
    """Returns the counter-clockwise angle span from `start` to `end` in degrees.

    Returns the angle span in the range of [0, 360], 360 is a full circle.
    Full circle handling is a special case, because normalization of angles
    which describe a full circle would return 0 if treated as regular angles.
    e.g. (0, 360) → 360, (0, -360) → 360, (180, -180) → 360.
    Input angles with the same value always return 0 by definition: (0, 0) → 0,
    (-180, -180) → 0, (360, 360) → 0.

    """
    # Input values are equal, returns 0 by definition:
    if math.isclose(start, end, abs_tol=DEG_ABS_TOL):
        return 0.0

    # Normalized start- and end angles are equal, but input values are
    # different, returns 360 by definition:
    start %= 360.0
    if math.isclose(start, end % 360.0, abs_tol=DEG_ABS_TOL):
        return 360.0

    # Special treatment for end angle == 360 deg:
    if not math.isclose(end, 360.0, abs_tol=DEG_ABS_TOL):
        end %= 360.0

    if end < start:
        end += 360.0
    return end - start


def arc_angle_span_rad(start: float, end: float) -> float:
    """Returns the counter-clockwise angle span from `start` to `end` in radians.

    Returns the angle span in the range of [0, 2π], 2π is a full circle.
    Full circle handling is a special case, because normalization of angles
    which describe a full circle would return 0 if treated as regular angles.
    e.g. (0, 2π) → 2π, (0, -2π) → 2π, (π, -π) → 2π.
    Input angles with the same value always return 0 by definition: (0, 0) → 0,
    (-π, -π) → 0, (2π, 2π) → 0.

    """
    tau = math.tau
    # Input values are equal, returns 0 by definition:
    if math.isclose(start, end, abs_tol=RAD_ABS_TOL):
        return 0.0

    # Normalized start- and end angles are equal, but input values are
    # different, returns 360 by definition:
    start %= tau
    if math.isclose(start, end % tau, abs_tol=RAD_ABS_TOL):
        return tau

    # Special treatment for end angle == 2π:
    if not math.isclose(end, tau, abs_tol=RAD_ABS_TOL):
        end %= tau

    if end < start:
        end += tau
    return end - start


def is_point_in_polygon_2d(
    point: Vec2, polygon: list[Vec2], abs_tol=TOLERANCE
) -> int:
    """
    Test if `point` is inside `polygon`.  Returns +1 for inside, 0 for on the 
    boundary and  -1 for outside.

    Supports convex and concave polygons with clockwise or counter-clockwise oriented
    polygon vertices.  Does not raise an exception for degenerated polygons.


    Args:
        point: 2D point to test as :class:`Vec2`
        polygon: list of 2D points as :class:`Vec2`
        abs_tol: tolerance for distance check

    Returns:
        +1 for inside, 0 for on the boundary, -1 for outside

    """
    # Source: http://www.faqs.org/faqs/graphics/algorithms-faq/
    # Subject 2.03: How do I find if a point lies within a polygon?
    # polygon: the Cython implementation needs a list as input to be fast!
    assert isinstance(polygon, list)
    if len(polygon) < 3: # empty polygon
        return -1

    if polygon[0].isclose(polygon[-1]):  # open polygon is required
        polygon = polygon[:-1]
    if len(polygon) < 3:
        return -1
    x = point.x
    y = point.y
    inside = False
    x1, y1 = polygon[-1]
    for x2, y2 in polygon:
        # is point on polygon boundary line:
        # is point in x-range of line
        a, b = (x2, x1) if x2 < x1 else (x1, x2)
        if a <= x <= b:
            # is point in y-range of line
            c, d = (y2, y1) if y2 < y1 else (y1, y2)
            if (c <= y <= d) and abs(
                (y2 - y1) * x - (x2 - x1) * y + (x2 * y1 - y2 * x1)
            ) <= abs_tol:
                return 0  # on boundary line
        if ((y1 <= y < y2) or (y2 <= y < y1)) and (
            x < (x2 - x1) * (y - y1) / (y2 - y1) + x1
        ):
            inside = not inside
        x1 = x2
        y1 = y2
    if inside:
        return 1  # inside polygon
    else:
        return -1  # outside polygon


# Values stored in GeoData RSS tag are not precise enough to match
# control calculation at epsg.io:
# Semi Major Axis: 6.37814e+06
# Semi Minor Axis: 6.35675e+06
WGS84_SEMI_MAJOR_AXIS = 6378137
WGS84_SEMI_MINOR_AXIS = 6356752.3142
WGS84_ELLIPSOID_ECCENTRIC = 0.08181919092890624
# WGS84_ELLIPSOID_ECCENTRIC = math.sqrt(
#    1.0 - WGS84_SEMI_MINOR_AXIS**2 / WGS84_SEMI_MAJOR_AXIS**2
# )
CONST_E2 = 1.3591409142295225  # math.e / 2.0
CONST_PI_2 = 1.5707963267948966  # math.pi / 2.0
CONST_PI_4 = 0.7853981633974483  # math.pi / 4.0


def gps_to_world_mercator(longitude: float, latitude: float) -> tuple[float, float]:
    """Transform GPS (long/lat) to World Mercator.
    
    Transform WGS84 `EPSG:4326 <https://epsg.io/4326>`_ location given as
    latitude and longitude in decimal degrees as used by GPS into World Mercator
    cartesian 2D coordinates in meters `EPSG:3395 <https://epsg.io/3395>`_.

    Args:
        longitude: represents the longitude value (East-West) in decimal degrees 
        latitude: represents the latitude value (North-South) in decimal degrees.
    
    .. versionadded:: 1.3.0

    """
    # From: https://epsg.io/4326
    # EPSG:4326 WGS84 - World Geodetic System 1984, used in GPS
    # To: https://epsg.io/3395
    # EPSG:3395 - World Mercator
    # Source: https://gis.stackexchange.com/questions/259121/transformation-functions-for-epsg3395-projection-vs-epsg3857
    longitude = math.radians(longitude)  # east
    latitude = math.radians(latitude)  # north
    a = WGS84_SEMI_MAJOR_AXIS
    e = WGS84_ELLIPSOID_ECCENTRIC
    e_sin_lat = math.sin(latitude) * e
    c = math.pow((1.0 - e_sin_lat) / (1.0 + e_sin_lat), e / 2.0)  # 7-7 p.44
    y = a * math.log(math.tan(CONST_PI_4 + latitude / 2.0) * c)  # 7-7 p.44
    x = a * longitude
    return x, y


def world_mercator_to_gps(x: float, y: float, tol: float = 1e-6) -> tuple[float, float]:
    """Transform World Mercator to GPS.

    Transform WGS84 World Mercator `EPSG:3395 <https://epsg.io/3395>`_
    location given as cartesian 2D coordinates x, y in meters into WGS84 decimal
    degrees as longitude and latitude `EPSG:4326 <https://epsg.io/4326>`_ as
    used by GPS.

    Args:
        x: coordinate WGS84 World Mercator
        y: coordinate WGS84 World Mercator
        tol: accuracy for latitude calculation

    .. versionadded:: 1.3.0

    """
    # From: https://epsg.io/3395
    # EPSG:3395 - World Mercator
    # To: https://epsg.io/4326
    # EPSG:4326 WGS84 - World Geodetic System 1984, used in GPS
    # Source: Map Projections - A Working Manual
    # https://pubs.usgs.gov/pp/1395/report.pdf
    a = WGS84_SEMI_MAJOR_AXIS
    e = WGS84_ELLIPSOID_ECCENTRIC
    e2 = e / 2.0
    pi2 = CONST_PI_2
    t = math.e ** (-y / a)  # 7-10 p.44
    latitude_ = pi2 - 2.0 * math.atan(t)  # 7-11 p.45
    while True:
        e_sin_lat = math.sin(latitude_) * e
        latitude = pi2 - 2.0 * math.atan(
            t * ((1.0 - e_sin_lat) / (1.0 + e_sin_lat)) ** e2
        )  # 7-9 p.44
        if abs(latitude - latitude_) < tol:
            break
        latitude_ = latitude

    longitude = x / a  # 7-12 p.45
    return math.degrees(longitude), math.degrees(latitude)



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest
from copy import deepcopy
from ezdxf.math import Vec3
from ezdxf.entities.dxfentity import (
    base_class,
    DXFAttributes,
    DXFNamespace,
    SubclassProcessor,
)
from ezdxf.lldxf.attributes import (
    group_code_mapping,
    DefSubclass,
    DXFAttr,
    XType,
)
from ezdxf.entities.dxfgfx import acdb_entity
from ezdxf.entities.line import acdb_line
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.lldxf.const import DXFAttributeError
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.lldxf.tags import Tags, DXFTag


class DXFEntity:
    """Mockup"""

    DXFTYPE = "DXFENTITY"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_line)


@pytest.fixture
def entity():
    return DXFEntity()


@pytest.fixture
def processor():
    return SubclassProcessor(ExtendedTags.from_text(TEST_1))


def test_handle_and_owner(entity, processor):
    attribs = DXFNamespace(processor, entity)
    assert attribs.handle == "FFFF"
    assert attribs.owner == "ABBA"
    assert attribs._entity is entity


def test_default_values(entity, processor):
    attribs = DXFNamespace(processor, entity)
    assert attribs.layer == "0"
    assert attribs.color == 256
    assert attribs.linetype == "BYLAYER"
    # this attributes do not really exist
    assert attribs.hasattr("layer") is False
    assert attribs.hasattr("color") is False
    assert attribs.hasattr("linetype") is False


def test_get_value_with_default(entity, processor):
    attribs = DXFNamespace(processor, entity)
    # return existing values
    assert attribs.get("handle", "0") == "FFFF"
    # return given default value not DXF default value, which would be '0'
    assert attribs.get("layer", "mozman") == "mozman"
    # attribute has to a valid DXF attribute
    with pytest.raises(DXFAttributeError):
        _ = attribs.get("hallo", 0)

    # attribs without default returns None -> will not exported to DXF file
    assert attribs.color_name is None


def test_set_values(entity, processor):
    attribs = DXFNamespace(processor, entity)
    attribs.handle = "CDEF"
    assert attribs.handle == "CDEF"
    attribs.set("owner", "DADA")
    assert attribs.owner == "DADA"
    # set new attribute
    attribs.color = 7
    assert attribs.color == 7
    attribs.set("linetype", "DOT")
    assert attribs.linetype == "DOT"
    # attribute has to a valid DXF attribute
    with pytest.raises(DXFAttributeError):
        attribs.hallo = 0
    with pytest.raises(DXFAttributeError):
        attribs.set("hallo", 0)


def test_value_types(entity, processor):
    attribs = DXFNamespace(processor, entity)
    attribs.handle = (
        None  # None is always accepted, attribute is ignored at export
    )
    assert attribs.handle is None
    attribs.handle = "XYZ"
    assert attribs.handle == "XYZ", "handle is just a string"
    attribs.handle = 123
    assert attribs.handle == "123", "handle is just a string"
    with pytest.raises(ValueError):
        attribs.color = "xxx"

    attribs.start = (1, 2, 3)  # type: Vec3
    assert attribs.start == (1, 2, 3)
    assert attribs.start.x == 1
    assert attribs.start.y == 2
    assert attribs.start.z == 3


def test_delete_attribs(entity, processor):
    attribs = DXFNamespace(processor, entity)
    attribs.layer = "mozman"
    assert attribs.layer == "mozman"
    del attribs.layer

    # default value
    assert attribs.layer == "0"
    with pytest.raises(DXFAttributeError):
        del attribs.color
    attribs.discard("color")  # delete silently if not exist
    with pytest.raises(DXFAttributeError):
        del attribs.unsupported_attribute


def test_is_supported(entity, processor):
    attribs = DXFNamespace(processor, entity)
    assert attribs.is_supported("linetype") is True
    assert (
        attribs.is_supported("true_color") is True
    )  # ezdxf does not care about DXF versions at runtime
    assert attribs.is_supported("xxx_mozman_xxx") is False


def test_dxftype(entity, processor):
    attribs = DXFNamespace(processor, entity)
    assert attribs.dxftype == "DXFENTITY"


def test_cloning(entity, processor):
    attribs = DXFNamespace(processor, entity)
    attribs.color = 77
    attribs2 = attribs.copy(entity)
    # clone everything
    assert attribs2._entity is attribs._entity
    assert attribs2.handle is attribs.handle
    assert attribs2.owner is attribs.owner
    assert attribs2.color == 77
    # do not harm original entity
    assert attribs._entity is entity
    assert attribs.handle == "FFFF"
    assert attribs.owner == "ABBA"
    # change clone
    attribs2.color = 13
    assert attribs.color == 77
    assert attribs2.color == 13


def test_deepcopy_usage(entity, processor):
    attribs = DXFNamespace(processor, entity)
    attribs.color = 77

    attribs2 = deepcopy(attribs)
    # clone everything
    assert attribs2._entity is attribs._entity
    assert attribs2.handle is attribs.handle
    assert attribs2.owner is attribs.owner
    assert attribs2.color == 77
    # do not harm original entity
    assert attribs._entity is entity
    assert attribs.handle == "FFFF"
    assert attribs.owner == "ABBA"
    # change clone
    attribs2.color = 13
    assert attribs.color == 77
    assert attribs2.color == 13


def test_dxf_export_one_attribute(entity, processor):
    attribs = DXFNamespace(processor, entity)
    tagwriter = TagCollector()
    attribs.export_dxf_attribs(tagwriter, "handle")
    assert len(tagwriter.tags) == 1
    assert tagwriter.tags[0] == (5, "FFFF")
    with pytest.raises(DXFAttributeError):
        attribs.export_dxf_attribute(tagwriter, "mozman")


def test_dxf_export_two_attribute(entity, processor):
    attribs = DXFNamespace(processor, entity)
    tagwriter = TagCollector()
    attribs.export_dxf_attribs(tagwriter, ["handle", "owner"])
    assert len(tagwriter.tags) == 2
    assert tagwriter.tags[0] == (5, "FFFF")
    assert tagwriter.tags[1] == (330, "ABBA")


class TestUpdateDXFAttributes:
    """Tests for DXFNamespace.update()"""

    @pytest.fixture
    def dxf(self, processor, entity):
        ns = DXFNamespace(processor, entity)
        ns.color = 7
        ns.handle = "ABBA"
        ns.owner = "FEFE"
        return ns

    def test_regular_usage(self, dxf):
        dxf.update({"color": 1})
        assert dxf.color == 1

    def test_protect_entity_back_link_from_update(self, dxf):
        entity = dxf._entity
        dxf.update({"_entity": None})
        assert entity is dxf._entity

    def test_protect_handle_and_owner_from_update_by_default(self, dxf):
        dxf.update({"handle": "BAAB", "owner": "DEDE"})
        assert dxf.handle == "ABBA"
        assert dxf.owner == "FEFE"

    def test_can_exclude_attributes_from_update(self, dxf):
        entity = dxf._entity
        dxf.update({"color": 1, "_entity": None}, exclude={"color"})
        assert dxf.color == 7
        assert (
            dxf._entity is entity
        ), "entity back-link has to be protected from update"

    def test_can_update_handle_and_owner_on_demand(self, dxf):
        entity = dxf._entity
        # do not include "handle" and "owner" in custom exclude set:
        exclude = set()
        dxf.update(
            {"handle": "BAAB", "owner": "DEDE", "_entity": None},
            exclude=exclude,
        )
        assert dxf.handle == "BAAB"
        assert dxf.owner == "DEDE"
        assert (
            dxf._entity is entity
        ), "entity back-link always has to be protected from update"

    def test_raises_attribute_error_for_invalid_attribute(self, dxf):
        with pytest.raises(AttributeError):
            dxf.update({"xxx": "yyy"})

    def test_raises_value_error_for_invalid_attribute_value(self, dxf):
        with pytest.raises(ValueError):
            dxf.update({"color": "yyy"})

    def test_update_can_ignore_errors_on_demand(self, dxf):
        dxf.update({"xxx": "yyy", "color": "zzz"}, ignore_errors=True)
        assert dxf.hasattr("xxx") is False
        assert dxf.color == 7


@pytest.fixture
def subclass():
    return DefSubclass(
        "AcDbTest",
        {
            "test1": DXFAttr(1),
            "test2": DXFAttr(2),
            "test3": DXFAttr(1),  # duplicate group code
            "callback": DXFAttr(3, xtype=XType.callback),
            "none_callback": DXFAttr(3),  # duplicate group code
        },
    )


@pytest.fixture
def cls(subclass):
    class TestEntity(DXFEntity):
        DXFATTRIBS = DXFAttributes(subclass)

    return TestEntity


def load_tags_fast(cls, subclass, data):
    ns = DXFNamespace(entity=cls())
    mapping = group_code_mapping(subclass)
    proc = SubclassProcessor(ExtendedTags(tags=data))
    unprocessed_tags = proc.fast_load_dxfattribs(ns, mapping, 0)
    return ns, unprocessed_tags


def test_if_fast_load_handles_duplicate_group_codes(cls, subclass):
    data = Tags(
        [
            DXFTag(0, "ENTITY"),  # First subclass tag should be ignored
            DXFTag(1, "1"),  # duplicate group code
            DXFTag(2, "2"),
            DXFTag(1, "3"),  # duplicate group code
            DXFTag(7, "unprocessed tag"),
        ]
    )
    ns, unprocessed_tags = load_tags_fast(cls, subclass, data)
    assert ns.test1 == "1"
    assert ns.test2 == "2"
    assert ns.test3 == "3"
    assert len(unprocessed_tags) == 1
    assert unprocessed_tags[0] == (7, "unprocessed tag")


def test_if_fast_load_handles_callback_group_codes(cls, subclass):
    data = Tags(
        [
            DXFTag(0, "ENTITY"),  # First subclass tag should be ignored
            DXFTag(3, "X"),  # callback value
            DXFTag(3, "Y"),  # none callback value
        ]
    )
    ns, unprocessed_tags = load_tags_fast(cls, subclass, data)
    assert ns.none_callback == "Y"
    assert ns.hasattr("callback") is False


def test_if_fast_load_handles_unprocessed_duplicate_group_codes(cls, subclass):
    data = Tags(
        [
            DXFTag(0, "ENTITY"),  # First subclass tag should be ignored
            DXFTag(1, "1"),  # duplicate group code
            DXFTag(1, "3"),  # duplicate group code
            DXFTag(
                1, "5"
            ),  # duplicate group code, but without an attribute definition
        ]
    )
    ns, unprocessed_tags = load_tags_fast(cls, subclass, data)
    assert ns.test1 == "1"
    assert ns.test3 == "3"
    # only two group code 1 attributes are defined:
    assert len(unprocessed_tags) == 1
    assert unprocessed_tags[0] == (1, "5")


def test_detect_implementation_version_returns_stored_version():
    processor = SubclassProcessor(
        ExtendedTags.from_text(GEODATA_V1), dxfversion="AC1015"
    )
    assert (
        processor.detect_implementation_version(
            subclass_index=1, group_code=90, default=2
        )
        == 1
    )


def test_detect_implementation_version_returns_default():
    processor = SubclassProcessor(
        ExtendedTags.from_text(TEST_1), dxfversion="AC1015"
    )
    assert (
        processor.detect_implementation_version(
            subclass_index=1, group_code=90, default=2
        )
        == 2
    )


TEST_1 = """0
DXFENTITY
5
FFFF
330
ABBA
"""

GEODATA_V1 = """0
GEODATA
5
395
330
1C5
100
AcDbGeoData
90
1
70
0
"""
if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Iterable, cast
from pathlib import Path
import pytest
import ezdxf
from ezdxf import recover
from ezdxf.entities import MText

DATA = Path(__file__).parent / "data"
TEXT_FRAME = "mtext_text_frame.dxf"


def load_mtext_entities(name: str):
    doc = ezdxf.readfile(DATA / name)
    return doc


def recover_mtext_entities(name: str):
    doc, auditor = recover.readfile(DATA / name)
    return doc


@pytest.fixture(scope="module", params=["load", "recover"])
def doc(request):
    if request.param == "load":
        return load_mtext_entities(TEXT_FRAME)
    elif request.param == "recover":
        return recover_mtext_entities(TEXT_FRAME)


def test_remove_mtext_text_frame_at_loading_stage(doc):
    msp = doc.modelspace()
    assert (
        len(msp.query("LWPOLYLINE")) == 0
    ), "text frame entities should be removed"
    mtext = msp.query("MTEXT").first
    assert bool(mtext.dxf.bg_fill & 16) is True


def test_remove_column_text_frames_at_loading_stage():
    doc = ezdxf.readfile(DATA / "mtext_framed_columns.dxf")
    msp = doc.modelspace()
    assert (
        len(msp.query("LWPOLYLINE")) == 0
    ), "text frame entities should be removed"


def test_do_not_export_mtext_text_borders(tmp_path):
    doc1 = ezdxf.readfile(DATA / TEXT_FRAME)
    doc1.saveas(tmp_path / TEXT_FRAME)
    doc2 = ezdxf.readfile(tmp_path / TEXT_FRAME)
    msp = doc2.modelspace()
    assert (
        len(msp.query("LWPOLYLINE")) == 0
    ), "text frame entities should be removed"
    mtext = msp.query("MTEXT").first
    assert mtext.xdata is None, "XDATA for MTEXT_TEXT_BORDERS should be removed"
    assert bool(mtext.dxf.bg_fill & 16) is True


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entities import DXFGraphic


def test_set_hyperlink():
    entity = DXFGraphic()
    assert entity.has_hyperlink() is False
    entity.set_hyperlink("link")
    assert entity.has_hyperlink() is True
    hyperlink, description, location = entity.get_hyperlink()
    assert hyperlink == "link"
    assert description == ""
    assert location == ""


def test_set_description():
    entity = DXFGraphic()
    entity.set_hyperlink("link", "description")
    hyperlink, description, location = entity.get_hyperlink()
    assert hyperlink == "link"
    assert description == "description"
    assert location == ""


def test_set_location():
    entity = DXFGraphic()
    entity.set_hyperlink("link", "description", "location")
    hyperlink, description, location = entity.get_hyperlink()
    assert hyperlink == "link"
    assert description == "description"
    assert location == "location"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2008, Casey Duncan (casey dot duncan at gmail dot com)
# see LICENSE.txt for details
"""Noise functions for procedural generation of content

Contains native code implementations of Perlin improved noise (with
fBm capabilities) and Perlin simplex noise. Also contains a fast
"fake noise" implementation in GLSL for execution in shaders.

Copyright (c) 2008, Casey Duncan (casey dot duncan at gmail dot com)
"""
__version__ = "1.2.1"

from math import floor, fmod, sqrt
from random import randint

# 3D Gradient vectors
# fmt: off
_GRAD3 = (
    (1, 1, 0), (-1, 1, 0), (1, -1, 0), (-1, -1, 0),
    (1, 0, 1), (-1, 0, 1), (1, 0, -1), (-1, 0, -1),
    (0, 1, 1), (0, -1, 1), (0, 1, -1), (0, -1, -1),
    (1, 1, 0), (0, -1, 1), (-1, 1, 0), (0, -1, -1),
)
# fmt: on

# 4D Gradient vectors
# fmt: off
_GRAD4 = (
    (0, 1, 1, 1), (0, 1, 1, -1), (0, 1, -1, 1), (0, 1, -1, -1),
    (0, -1, 1, 1), (0, -1, 1, -1), (0, -1, -1, 1), (0, -1, -1, -1),
    (1, 0, 1, 1), (1, 0, 1, -1), (1, 0, -1, 1), (1, 0, -1, -1),
    (-1, 0, 1, 1), (-1, 0, 1, -1), (-1, 0, -1, 1), (-1, 0, -1, -1),
    (1, 1, 0, 1), (1, 1, 0, -1), (1, -1, 0, 1), (1, -1, 0, -1),
    (-1, 1, 0, 1), (-1, 1, 0, -1), (-1, -1, 0, 1), (-1, -1, 0, -1),
    (1, 1, 1, 0), (1, 1, -1, 0), (1, -1, 1, 0), (1, -1, -1, 0),
    (-1, 1, 1, 0), (-1, 1, -1, 0), (-1, -1, 1, 0), (-1, -1, -1, 0)
)
# fmt: on

# A lookup table to traverse the simplex around a given point in 4D. 
# Details can be found where this table is used, in the 4D noise method.
# fmt: off
_SIMPLEX = (
    (0, 1, 2, 3), (0, 1, 3, 2), (0, 0, 0, 0), (0, 2, 3, 1), (0, 0, 0, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (1, 2, 3, 0), (0, 2, 1, 3), (0, 0, 0, 0),
    (0, 3, 1, 2), (0, 3, 2, 1), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0),
    (1, 3, 2, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (1, 2, 0, 3),
    (0, 0, 0, 0), (1, 3, 0, 2), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0),
    (2, 3, 0, 1), (2, 3, 1, 0), (1, 0, 2, 3), (1, 0, 3, 2), (0, 0, 0, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (2, 0, 3, 1), (0, 0, 0, 0), (2, 1, 3, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (2, 0, 1, 3), (0, 0, 0, 0),
    (0, 0, 0, 0), (0, 0, 0, 0), (3, 0, 1, 2), (3, 0, 2, 1), (0, 0, 0, 0),
    (3, 1, 2, 0), (2, 1, 0, 3), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0),
    (3, 1, 0, 2), (0, 0, 0, 0), (3, 2, 0, 1), (3, 2, 1, 0)
)
# fmt: on
# Simplex skew constants
_F2 = 0.5 * (sqrt(3.0) - 1.0)
_G2 = (3.0 - sqrt(3.0)) / 6.0
_F3 = 1.0 / 3.0
_G3 = 1.0 / 6.0

# fmt: off
_permutation = (
    151, 160, 137, 91, 90, 15, 131, 13, 201, 95, 96, 53, 194, 233, 7, 225,
    140, 36, 103, 30, 69, 142, 8, 99, 37, 240, 21, 10, 23, 190, 6, 148, 247,
    120, 234, 75, 0, 26, 197, 62, 94, 252, 219, 203, 117, 35, 11, 32, 57,
    177, 33, 88, 237, 149, 56, 87, 174, 20, 125, 136, 171, 168, 68, 175, 74,
    165, 71, 134, 139, 48, 27, 166, 77, 146, 158, 231, 83, 111, 229, 122,
    60, 211, 133, 230, 220, 105, 92, 41, 55, 46, 245, 40, 244, 102, 143, 54,
    65, 25, 63, 161, 1, 216, 80, 73, 209, 76, 132, 187, 208, 89, 18, 169,
    200, 196, 135, 130, 116, 188, 159, 86, 164, 100, 109, 198, 173, 186, 3,
    64, 52, 217, 226, 250, 124, 123, 5, 202, 38, 147, 118, 126, 255, 82, 85,
    212, 207, 206, 59, 227, 47, 16, 58, 17, 182, 189, 28, 42, 223, 183, 170,
    213, 119, 248, 152, 2, 44, 154, 163, 70, 221, 153, 101, 155, 167, 43,
    172, 9, 129, 22, 39, 253, 9, 98, 108, 110, 79, 113, 224, 232, 178, 185,
    112, 104, 218, 246, 97, 228, 251, 34, 242, 193, 238, 210, 144, 12, 191,
    179, 162, 241, 81, 51, 145, 235, 249, 14, 239, 107, 49, 192, 214, 31,
    181, 199, 106, 157, 184, 84, 204, 176, 115, 121, 50, 45, 127, 4, 150,
    254, 138, 236, 205, 93, 222, 114, 67, 29, 24, 72, 243, 141, 128, 195,
    78, 66, 215, 61, 156, 180
)
# fmt: on


class BaseNoise:
    """Noise abstract base class"""
    period = len(_permutation)
    # Double permutation array so we don't need to wrap
    permutation = _permutation * 2

    def __init__(self, period=None, permutation_table=None):
        """Initialize the noise generator. With no arguments, the default
        period and permutation table are used (256). The default permutation
        table generates the exact same noise pattern each time.

        An integer period can be specified, to generate a random permutation
        table with period elements. The period determines the (integer)
        interval that the noise repeats, which is useful for creating tiled
        textures.  period should be a power-of-two, though this is not
        enforced. Note that the speed of the noise algorithm is independent of
        the period size, though larger periods mean a larger table, which
        consume more memory.

        A permutation table consisting of an iterable sequence of whole
        numbers can be specified directly. This should have a power-of-two
        length. Typical permutation tables are a sequence of unique integers in
        the range [0,period) in random order, though other arrangements could
        prove useful, they will not be "pure" simplex noise. The largest
        element in the sequence must be no larger than period-1.

        period and permutation_table may not be specified together.
        """
        if period is not None and permutation_table is not None:
            raise ValueError(
                "Can specify either period or permutation_table, not both"
            )
        if period is not None:
            self.randomize(period)
        elif permutation_table is not None:
            self.permutation = tuple(permutation_table) * 2
            self.period = len(permutation_table)

    def randomize(self, period=None):
        """Randomize the permutation table used by the noise functions. This
        makes them generate a different noise pattern for the same inputs.
        """
        if period is not None:
            self.period = period
        perm = list(range(self.period))
        perm_right = self.period - 1
        for i in list(perm):
            j = randint(0, perm_right)
            perm[i], perm[j] = perm[j], perm[i]
        self.permutation = tuple(perm) * 2


class SimplexNoise(BaseNoise):
    """Perlin simplex noise generator

    Adapted from Stefan Gustavson's Java implementation described here:

    http://staffwww.itn.liu.se/~stegu/simplexnoise/simplexnoise.pdf

    To summarize:

    "In 2001, Ken Perlin presented 'simplex noise', a replacement for his classic
    noise algorithm.  Classic 'Perlin noise' won him an academy award and has
    become an ubiquitous procedural primitive for computer graphics over the
    years, but in hindsight it has quite a few limitations.  Ken Perlin himself
    designed simplex noise specifically to overcome those limitations, and he
    spent a lot of good thinking on it. Therefore, it is a better idea than his
    original algorithm. A few of the more prominent advantages are:

    * Simplex noise has a lower computational complexity and requires fewer
      multiplications.
    * Simplex noise scales to higher dimensions (4D, 5D and up) with much less
      computational cost, the complexity is O(N) for N dimensions instead of
      the O(2^N) of classic Noise.
    * Simplex noise has no noticeable directional artifacts.  Simplex noise has
      a well-defined and continuous gradient everywhere that can be computed
      quite cheaply.
    * Simplex noise is easy to implement in hardware."
    """

    def noise2(self, x, y):
        """2D Perlin simplex noise.

        Return a floating point value from -1 to 1 for the given x, y coordinate.
        The same value is always returned for a given x, y pair unless the
        permutation table changes (see randomize above).
        """
        # Skew input space to determine which simplex (triangle) we are in
        s = (x + y) * _F2
        i = floor(x + s)
        j = floor(y + s)
        t = (i + j) * _G2
        x0 = x - (i - t)  # "Unskewed" distances from cell origin
        y0 = y - (j - t)

        if x0 > y0:
            i1 = 1
            j1 = 0  # Lower triangle, XY order: (0,0)->(1,0)->(1,1)
        else:
            i1 = 0
            j1 = 1  # Upper triangle, YX order: (0,0)->(0,1)->(1,1)

        x1 = x0 - i1 + _G2  # Offsets for middle corner in (x,y) unskewed coords
        y1 = y0 - j1 + _G2
        x2 = (
            x0 + _G2 * 2.0 - 1.0
        )  # Offsets for last corner in (x,y) unskewed coords
        y2 = y0 + _G2 * 2.0 - 1.0

        # Determine hashed gradient indices of the three simplex corners
        perm = self.permutation
        ii = int(i) % self.period
        jj = int(j) % self.period
        gi0 = perm[ii + perm[jj]] % 12
        gi1 = perm[ii + i1 + perm[jj + j1]] % 12
        gi2 = perm[ii + 1 + perm[jj + 1]] % 12

        # Calculate the contribution from the three corners
        tt = 0.5 - x0 ** 2 - y0 ** 2
        if tt > 0:
            g = _GRAD3[gi0]
            noise = tt ** 4 * (g[0] * x0 + g[1] * y0)
        else:
            noise = 0.0

        tt = 0.5 - x1 ** 2 - y1 ** 2
        if tt > 0:
            g = _GRAD3[gi1]
            noise += tt ** 4 * (g[0] * x1 + g[1] * y1)

        tt = 0.5 - x2 ** 2 - y2 ** 2
        if tt > 0:
            g = _GRAD3[gi2]
            noise += tt ** 4 * (g[0] * x2 + g[1] * y2)

        return noise * 70.0  # scale noise to [-1, 1]

    def noise3(self, x, y, z):
        """3D Perlin simplex noise.

        Return a floating point value from -1 to 1 for the given x, y, z coordinate.
        The same value is always returned for a given x, y, z pair unless the
        permutation table changes (see randomize above).
        """
        # Skew the input space to determine which simplex cell we're in
        s = (x + y + z) * _F3
        i = floor(x + s)
        j = floor(y + s)
        k = floor(z + s)
        t = (i + j + k) * _G3
        x0 = x - (i - t)  # "Unskewed" distances from cell origin
        y0 = y - (j - t)
        z0 = z - (k - t)

        # For the 3D case, the simplex shape is a slightly irregular tetrahedron.
        # Determine which simplex we are in.
        if x0 >= y0:
            if y0 >= z0:
                i1 = 1
                j1 = 0
                k1 = 0
                i2 = 1
                j2 = 1
                k2 = 0
            elif x0 >= z0:
                i1 = 1
                j1 = 0
                k1 = 0
                i2 = 1
                j2 = 0
                k2 = 1
            else:
                i1 = 0
                j1 = 0
                k1 = 1
                i2 = 1
                j2 = 0
                k2 = 1
        else:  # x0 < y0
            if y0 < z0:
                i1 = 0
                j1 = 0
                k1 = 1
                i2 = 0
                j2 = 1
                k2 = 1
            elif x0 < z0:
                i1 = 0
                j1 = 1
                k1 = 0
                i2 = 0
                j2 = 1
                k2 = 1
            else:
                i1 = 0
                j1 = 1
                k1 = 0
                i2 = 1
                j2 = 1
                k2 = 0

        # Offsets for remaining corners
        x1 = x0 - i1 + _G3
        y1 = y0 - j1 + _G3
        z1 = z0 - k1 + _G3
        x2 = x0 - i2 + 2.0 * _G3
        y2 = y0 - j2 + 2.0 * _G3
        z2 = z0 - k2 + 2.0 * _G3
        x3 = x0 - 1.0 + 3.0 * _G3
        y3 = y0 - 1.0 + 3.0 * _G3
        z3 = z0 - 1.0 + 3.0 * _G3

        # Calculate the hashed gradient indices of the four simplex corners
        perm = self.permutation
        ii = int(i) % self.period
        jj = int(j) % self.period
        kk = int(k) % self.period
        gi0 = perm[ii + perm[jj + perm[kk]]] % 12
        gi1 = perm[ii + i1 + perm[jj + j1 + perm[kk + k1]]] % 12
        gi2 = perm[ii + i2 + perm[jj + j2 + perm[kk + k2]]] % 12
        gi3 = perm[ii + 1 + perm[jj + 1 + perm[kk + 1]]] % 12

        # Calculate the contribution from the four corners
        noise = 0.0
        tt = 0.6 - x0 ** 2 - y0 ** 2 - z0 ** 2
        if tt > 0:
            g = _GRAD3[gi0]
            noise = tt ** 4 * (g[0] * x0 + g[1] * y0 + g[2] * z0)
        else:
            noise = 0.0

        tt = 0.6 - x1 ** 2 - y1 ** 2 - z1 ** 2
        if tt > 0:
            g = _GRAD3[gi1]
            noise += tt ** 4 * (g[0] * x1 + g[1] * y1 + g[2] * z1)

        tt = 0.6 - x2 ** 2 - y2 ** 2 - z2 ** 2
        if tt > 0:
            g = _GRAD3[gi2]
            noise += tt ** 4 * (g[0] * x2 + g[1] * y2 + g[2] * z2)

        tt = 0.6 - x3 ** 2 - y3 ** 2 - z3 ** 2
        if tt > 0:
            g = _GRAD3[gi3]
            noise += tt ** 4 * (g[0] * x3 + g[1] * y3 + g[2] * z3)

        return noise * 32.0


def lerp(t, a, b):
    return a + t * (b - a)


def grad3(hash, x, y, z):
    g = _GRAD3[hash % 16]
    return x * g[0] + y * g[1] + z * g[2]


class TileableNoise(BaseNoise):
    """Tileable implementation of Perlin "improved" noise. This
    is based on the reference implementation published here:

    http://mrl.nyu.edu/~perlin/noise/
    """

    def noise3(self, x, y, z, repeat, base=0.0):
        """Tileable 3D noise.

        repeat specifies the integer interval in each dimension
        when the noise pattern repeats.

        base allows a different texture to be generated for
        the same repeat interval.
        """
        i = int(fmod(floor(x), repeat))
        j = int(fmod(floor(y), repeat))
        k = int(fmod(floor(z), repeat))
        ii = (i + 1) % repeat
        jj = (j + 1) % repeat
        kk = (k + 1) % repeat
        if base:
            i += base
            j += base
            k += base
            ii += base
            jj += base
            kk += base

        x -= floor(x)
        y -= floor(y)
        z -= floor(z)
        fx = x ** 3 * (x * (x * 6 - 15) + 10)
        fy = y ** 3 * (y * (y * 6 - 15) + 10)
        fz = z ** 3 * (z * (z * 6 - 15) + 10)

        perm = self.permutation
        A = perm[i]
        AA = perm[A + j]
        AB = perm[A + jj]
        B = perm[ii]
        BA = perm[B + j]
        BB = perm[B + jj]

        return lerp(
            fz,
            lerp(
                fy,
                lerp(
                    fx,
                    grad3(perm[AA + k], x, y, z),
                    grad3(perm[BA + k], x - 1, y, z),
                ),
                lerp(
                    fx,
                    grad3(perm[AB + k], x, y - 1, z),
                    grad3(perm[BB + k], x - 1, y - 1, z),
                ),
            ),
            lerp(
                fy,
                lerp(
                    fx,
                    grad3(perm[AA + kk], x, y, z - 1),
                    grad3(perm[BA + kk], x - 1, y, z - 1),
                ),
                lerp(
                    fx,
                    grad3(perm[AB + kk], x, y - 1, z - 1),
                    grad3(perm[BB + kk], x - 1, y - 1, z - 1),
                ),
            ),
        )


_simplex = SimplexNoise()
snoise2 = _simplex.noise2
snoise3 = _simplex.noise3
_tileable = TileableNoise()
tnoise3 = _tileable.noise3



# Copyright (C) 2019, Manfred Moitzi
# License: MIT License
# Created: 2019-02-24
import pytest
import ezdxf


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    return doc.modelspace()


def test_create_blockref(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    assert blockref.dxf.name == "TESTBLOCK"
    assert blockref.dxf.insert == (0.0, 0.0)


def test_blockref_attrib_get_flags(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    attrib = blockref.add_attrib("TAG1", "Text1", (0, 1))
    assert attrib.is_const is False
    assert attrib.is_invisible is False
    assert attrib.is_verify is False
    assert attrib.is_preset is False


def test_blockref_attrib_set_flags(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    attrib = blockref.add_attrib("TAG1", "Text1", (0, 1))

    assert attrib.is_const is False
    attrib.is_const = True
    assert attrib.is_const is True

    assert attrib.is_invisible is False
    attrib.is_invisible = True
    assert attrib.is_invisible is True

    assert attrib.is_verify is False
    attrib.is_verify = True
    assert attrib.is_verify is True

    assert attrib.is_preset is False
    attrib.is_preset = True
    assert attrib.is_preset is True


def test_blockref_add_new_attribs(msp):
    entity_count = len(msp)
    db = msp.entitydb
    db_count = len(db)
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.add_attrib("TEST", "text", (0, 0))
    assert len(msp) == entity_count + 1
    assert (
        len(msp.entitydb) == db_count + 3
    ), "expected INSERT, ATTRIB and SEQEND in db"

    assert blockref.attribs_follow == 1
    attrib = blockref.get_attrib("TEST")
    assert attrib.dxf.text == "text"


def test_blockref_add_to_attribs(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.add_attrib("TEST1", "text1", (0, 0))
    blockref.add_attrib("TEST2", "text2", (0, 0))
    assert ["TEST1", "TEST2"] == [attrib.dxf.tag for attrib in blockref.attribs]


def test_blockref_place(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.place(insert=(1, 2), scale=(0.5, 0.4, 0.3), rotation=37.0)
    assert blockref.dxf.insert == (1, 2)
    assert blockref.dxf.xscale == 0.5
    assert blockref.dxf.yscale == 0.4
    assert blockref.dxf.zscale == 0.3
    assert blockref.dxf.rotation == 37


def test_lockref_grid(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.grid(size=(2, 3), spacing=(5, 10))
    assert blockref.dxf.row_count == 2
    assert blockref.dxf.column_count == 3
    assert blockref.dxf.row_spacing == 5
    assert blockref.dxf.column_spacing == 10


def test_blockref_attribs_follow_abuse(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    assert len(blockref.attribs) == 0, "Attrib count should be 0"
    assert blockref.has_attrib("TEST") is False, "Attrib should not exists"

    blockref.add_attrib("TEST", "Text")
    assert len(blockref.attribs) == 1
    assert blockref.attribs_follow is True


def test_blockref_delete_not_existing_attrib(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    with pytest.raises(KeyError):
        blockref.delete_attrib("TEST")


def test_blockref_delete_not_existing_attrib_without_exception(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    # ignore=True, should ignore not existing ATTRIBS
    blockref.delete_attrib("TEST", ignore=True)
    assert True


def test_blockref_delete_last_attrib(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.add_attrib("TEST", "Text")
    assert blockref.has_attrib("TEST") is True

    # delete last attrib
    blockref.delete_attrib("TEST")
    assert len(blockref.attribs) == 0
    assert blockref.dxf.attribs_follow == 0


def test_blockref_delete_one_of_many_attribs(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.add_attrib("TEST0", "Text")
    blockref.add_attrib("TEST1", "Text")
    blockref.add_attrib("TEST2", "Text")
    assert len(blockref.attribs) == 3
    assert blockref.attribs_follow is True

    blockref.delete_attrib("TEST1")
    assert len(blockref.attribs) == 2
    assert blockref.attribs_follow is True

    assert blockref.has_attrib("TEST0") is True
    assert blockref.has_attrib("TEST1") is False
    assert blockref.has_attrib("TEST2") is True


def test_blockref_delete_first_of_many_attribs(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    blockref.add_attrib("TEST0", "Text")
    blockref.add_attrib("TEST1", "Text")
    assert len(blockref.attribs) == 2
    assert blockref.attribs_follow is True

    blockref.delete_attrib("TEST0")
    assert len(blockref.attribs) == 1
    assert blockref.attribs_follow is True

    assert blockref.has_attrib("TEST0") is False
    assert blockref.has_attrib("TEST1") is True


def test_blockref_delete_all_attribs(msp):
    blockref = msp.add_blockref("TESTBLOCK", (0, 0))
    # deleting none existing attribs, is ok
    blockref.delete_all_attribs()
    assert len(blockref.attribs) == 0
    assert blockref.dxf.attribs_follow == 0

    blockref.add_attrib("TEST0", "Text")
    blockref.add_attrib("TEST1", "Text")
    blockref.add_attrib("TEST2", "Text")
    assert len(blockref.attribs) == 3
    assert blockref.attribs_follow is True

    blockref.delete_all_attribs()
    assert len(blockref.attribs) == 0
    assert blockref.dxf.attribs_follow == 0



# Purpose: test dxfvalue
# Created: 12.03.2011
# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
from ezdxf.sections.header import HeaderVar


class TestDXFValue:
    def test_single_value_code(self):
        var = HeaderVar((0, "SECTION"))
        assert 0 == var.code

    def test_single_value_value(self):
        var = HeaderVar((0, "SECTION"))
        assert "SECTION" == var.value

    def test_single_value_str(self):
        var = HeaderVar((0, "SECTION"))
        assert "  0\nSECTION\n" == str(var)

    def test_not_ispoint(self):
        var = HeaderVar((0, "SECTION"))
        assert var.ispoint is False

    def test_ispoint(self):
        var = HeaderVar((10, (1, 1)))
        assert var.ispoint is True

    def test__point_2coords(self):
        var = HeaderVar((10, (1, 1)))
        assert (1, 1) == var.value

    def test_point_3coords(self):
        var = HeaderVar((10, (1, 2, 3)))
        assert (1, 2, 3) == var.value

    def test_point_str(self):
        var = HeaderVar((10, (1, 2, 3)))
        assert " 10\n1\n 20\n2\n 30\n3\n" == str(var)



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

import json
import ezdxf
from ezdxf.document import export_json_tags, load_json_tags
from ezdxf.lldxf.tagger import json_tag_loader

@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()

def test_export_json_tags(doc):
    result = export_json_tags(doc, compact=False)
    data = json.loads(result)

    assert isinstance(data, list)
    assert data[0] == [0, "SECTION"]
    assert data[1] == [2, "HEADER"]
    assert data[2] == [9, "$ACADVER"]
    assert data[3] == [1, doc.dxfversion]
    assert data[-1] == [0, "EOF"]


def test_export_compact_json_tags(doc):
    result = export_json_tags(doc, compact=True)
    data = json.loads(result)

    assert isinstance(data, list)
    assert data[0] == [0, "SECTION"]
    assert data[1] == [2, "HEADER"]
    assert data[12] == [9, "$INSBASE"]
    assert data[13] == [10, [0, 0, 0]]
    assert data[-1] == [0, "EOF"]


def test_load_json_tags(doc):
    data_str = export_json_tags(doc, compact=False)
    data = json.loads(data_str)
    tags = list(json_tag_loader(data))

    assert len(tags) == len(data)


def test_load_compact_json_tags():
    data = [[10, [1, 3, 5]]]
    tags = list(json_tag_loader(data))

    assert len(tags) == 3
    assert tags[0] == (10, 1)
    assert tags[1] == (20, 3)
    assert tags[2] == (30, 5)


def test_load_dxf_document_from_json_tags(doc):
    data_str = export_json_tags(doc, compact=False)
    data = json.loads(data_str)
    loaded_doc = load_json_tags(data)

    assert doc.dxfversion == loaded_doc.dxfversion


def test_load_dxf_document_from_compact_json_tags(doc):
    data_str = export_json_tags(doc, compact=True)
    data = json.loads(data_str)
    loaded_doc = load_json_tags(data)

    assert doc.dxfversion == loaded_doc.dxfversion


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_copy_simple_entity(doc):
    msp = doc.modelspace()
    psp = doc.layout("Layout1")
    circle = msp.add_circle(center=(2, 3), radius=1.5)
    assert circle.dxf.paperspace == 0
    len_msp = len(msp)
    len_psp = len(psp)
    new_circle = circle.copy_to_layout(psp)
    assert new_circle.dxf.paperspace == 1
    assert len_msp == len(msp)
    assert len_psp + 1 == len(psp)


def test_copy_polyline_entity(doc):
    msp = doc.modelspace()
    psp = doc.layout("Layout1")
    polyline = msp.add_polyline3d(points=[(1, 1, 1), (3, 2, -1), (7, 4, 4)])
    assert polyline.dxf.paperspace == 0
    len_msp = len(msp)
    len_psp = len(psp)
    new_polyline = polyline.copy_to_layout(psp)
    assert new_polyline.dxf.paperspace == 1
    assert len_msp == len(msp)
    assert len_psp + 1 == len(
        psp
    )  # just POLYLINE, attached entities do not appear in a layout space


def test_move_simple_entity(doc):
    msp = doc.modelspace()
    psp = doc.layout("Layout1")
    circle = msp.add_circle(center=(2, 3), radius=1.5)
    assert circle.dxf.paperspace == 0
    len_msp = len(msp)
    len_psp = len(psp)
    circle.move_to_layout(psp)
    assert circle.dxf.paperspace == 1
    assert len_msp - 1 == len(msp)
    assert len_psp + 1 == len(psp)


def test_move_polyline_to_paperspace(doc):
    msp = doc.modelspace()
    psp = doc.layout("Layout1")
    polyline = msp.add_polyline3d(points=[(1, 1, 1), (3, 2, -1), (7, 4, 4)])
    assert polyline.dxf.paperspace == 0
    len_msp = len(msp)
    len_psp = len(psp)
    polyline.move_to_layout(psp)
    assert polyline.dxf.paperspace == 1
    assert len_msp - 1 == len(msp)
    assert len_psp + 1 == len(psp)
    for vertex in polyline.vertices:
        assert vertex.dxf.paperspace == polyline.dxf.paperspace
        assert (
            vertex.dxf.owner == polyline.dxf.handle
        ), "POLYLINE is owner of VERTEX entities"


def test_move_polyline_to_block(doc):
    msp = doc.modelspace()
    block = doc.blocks.new("Test1")
    polyline = msp.add_polyline3d(points=[(1, 1, 1), (3, 2, -1), (7, 4, 4)])
    assert polyline.dxf.paperspace == 0
    len_msp = len(msp)
    len_block = len(block)
    polyline.move_to_layout(block)
    assert polyline.dxf.paperspace == 0
    assert len_msp - 1 == len(msp)
    assert len_block + 1 == len(block)
    for vertex in polyline.vertices:
        assert vertex.dxf.paperspace == polyline.dxf.paperspace
        assert (
            vertex.dxf.owner == polyline.dxf.handle
        ), "POLYLINE is owner of VERTEX entities"


def test_move_from_block_to_block(doc):
    source_block = doc.blocks.new("Test2")
    target_block = doc.blocks.new("Test3")
    polyline = source_block.add_polyline3d(points=[(1, 1, 1), (3, 2, -1), (7, 4, 4)])

    polyline.move_to_layout(target_block)
    assert len(source_block) == 0
    assert len(target_block) == 1
    # and back by layout.move_to_layout()
    target_block.move_to_layout(polyline, source_block)
    assert len(source_block) == 1
    assert len(target_block) == 0



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    NamedTuple,
    Any,
    Sequence,
    Iterator,
    Union,
    Iterable,
    cast,
    TYPE_CHECKING,
    List,
    Tuple,
    Optional,
)
from typing_extensions import TypeAlias
import math
import struct
from datetime import datetime

from ezdxf.math import Vec3
from . import const
from .const import ParsingError, Tags, InvalidLinkStructure
from .hdr import AcisHeader
from .abstract import (
    AbstractEntity,
    DataLoader,
    AbstractBuilder,
    DataExporter,
    EntityExporter,
)

if TYPE_CHECKING:
    from .entities import AcisEntity


class Token(NamedTuple):
    """Named tuple to store tagged value tokens of the SAB format."""

    tag: int
    value: Any

    def __str__(self):
        return f"(0x{self.tag:02x}, {str(self.value)})"


SabRecord: TypeAlias = List[Token]


class Decoder:
    def __init__(self, data: bytes):
        self.data = data
        self.index: int = 0

    @property
    def has_data(self) -> bool:
        return self.index < len(self.data)

    def read_header(self) -> AcisHeader:
        header = AcisHeader()
        for signature in const.SIGNATURES:
            if self.data.startswith(signature):
                self.index = len(signature)
                break
        else:
            raise ParsingError("not a SAB file")
        header.version = self.read_int()
        header.n_records = self.read_int()
        header.n_entities = self.read_int()
        header.flags = self.read_int()
        header.product_id = self.read_str_tag()
        header.acis_version = self.read_str_tag()
        date = self.read_str_tag()
        header.creation_date = datetime.strptime(date, const.DATE_FMT)
        header.units_in_mm = self.read_double_tag()
        # tolerances are ignored
        _ = self.read_double_tag()  # res_tol
        _ = self.read_double_tag()  # nor_tol
        return header

    def forward(self, count: int):
        pos = self.index
        self.index += count
        return pos

    def read_byte(self) -> int:
        pos = self.forward(1)
        return self.data[pos]

    def read_bytes(self, count: int) -> bytes:
        pos = self.forward(count)
        return self.data[pos : pos + count]

    def read_int(self) -> int:
        pos = self.forward(4)
        values = struct.unpack_from("<i", self.data, pos)[0]
        return values

    def read_float(self) -> float:
        pos = self.forward(8)
        return struct.unpack_from("<d", self.data, pos)[0]

    def read_floats(self, count: int) -> Sequence[float]:
        pos = self.forward(8 * count)
        return struct.unpack_from(f"<{count}d", self.data, pos)

    def read_str(self, length) -> str:
        text = self.read_bytes(length)
        return text.decode()

    def read_str_tag(self) -> str:
        tag = self.read_byte()
        if tag != Tags.STR:
            raise ParsingError("string tag (7) not found")
        return self.read_str(self.read_byte())

    def read_double_tag(self) -> float:
        tag = self.read_byte()
        if tag != Tags.DOUBLE:
            raise ParsingError("double tag (6) not found")
        return self.read_float()

    def read_record(self) -> SabRecord:
        def entity_name():
            return "-".join(entity_type)

        values: SabRecord = []
        entity_type: list[str] = []
        subtype_level: int = 0
        while True:
            if not self.has_data:
                if values:
                    token = values[0]
                    if token.value in const.DATA_END_MARKERS:
                        return values
                raise ParsingError("pre-mature end of data")
            tag = self.read_byte()
            if tag == Tags.INT:
                values.append(Token(tag, self.read_int()))
            elif tag == Tags.DOUBLE:
                values.append(Token(tag, self.read_float()))
            elif tag == Tags.STR:
                values.append(Token(tag, self.read_str(self.read_byte())))
            elif tag == Tags.POINTER:
                values.append(Token(tag, self.read_int()))
            elif tag == Tags.BOOL_TRUE:
                values.append(Token(tag, True))
            elif tag == Tags.BOOL_FALSE:
                values.append(Token(tag, False))
            elif tag == Tags.LITERAL_STR:
                values.append(Token(tag, self.read_str(self.read_int())))
            elif tag == Tags.ENTITY_TYPE_EX:
                entity_type.append(self.read_str(self.read_byte()))
            elif tag == Tags.ENTITY_TYPE:
                entity_type.append(self.read_str(self.read_byte()))
                values.append(Token(tag, entity_name()))
                entity_type.clear()
            elif tag == Tags.LOCATION_VEC:
                values.append(Token(tag, self.read_floats(3)))
            elif tag == Tags.DIRECTION_VEC:
                values.append(Token(tag, self.read_floats(3)))
            elif tag == Tags.ENUM:
                values.append(Token(tag, self.read_int()))
            elif tag == Tags.UNKNOWN_0x17:
                values.append(Token(tag, self.read_float()))
            elif tag == Tags.SUBTYPE_START:
                subtype_level += 1
                values.append(Token(tag, subtype_level))
            elif tag == Tags.SUBTYPE_END:
                values.append(Token(tag, subtype_level))
                subtype_level -= 1
            elif tag == Tags.RECORD_END:
                return values
            else:
                raise ParsingError(
                    f"unknown SAB tag: 0x{tag:x} ({tag}) in entity '{values[0].value}'"
                )

    def read_records(self) -> Iterator[SabRecord]:
        while True:
            try:
                if self.has_data:
                    yield self.read_record()
                else:
                    return
            except IndexError:
                return


class SabEntity(AbstractEntity):
    """Low level representation of an ACIS entity (node)."""

    def __init__(
        self,
        name: str,
        attr_ptr: int = -1,
        id: int = -1,
        data: Optional[SabRecord] = None,
    ):
        self.name = name
        self.attr_ptr = attr_ptr
        self.id = id
        self.data: SabRecord = data if data is not None else []
        self.attributes: "SabEntity" = None  # type: ignore

    def __str__(self):
        return f"{self.name}({self.id})"


NULL_PTR = SabEntity(const.NULL_PTR_NAME, -1, -1, tuple())  # type: ignore


class SabDataLoader(DataLoader):
    def __init__(self, data: SabRecord, version: int):
        self.version = version
        self.data = data
        self.index = 0

    def has_data(self) -> bool:
        return self.index <= len(self.data)

    def read_int(self, skip_sat: Optional[int] = None) -> int:
        token = self.data[self.index]
        if token.tag == Tags.INT:
            self.index += 1
            return cast(int, token.value)
        raise ParsingError(f"expected int token, got {token}")

    def read_double(self) -> float:
        token = self.data[self.index]
        if token.tag == Tags.DOUBLE:
            self.index += 1
            return cast(float, token.value)
        raise ParsingError(f"expected double token, got {token}")

    def read_interval(self) -> float:
        finite = self.read_bool("F", "I")
        if finite:
            return self.read_double()
        return math.inf

    def read_vec3(self) -> tuple[float, float, float]:
        token = self.data[self.index]
        if token.tag in (Tags.LOCATION_VEC, Tags.DIRECTION_VEC):
            self.index += 1
            return cast(Tuple[float, float, float], token.value)
        raise ParsingError(f"expected vector token, got {token}")

    def read_bool(self, true: str, false: str) -> bool:
        token = self.data[self.index]
        if token.tag == Tags.BOOL_TRUE:
            self.index += 1
            return True
        elif token.tag == Tags.BOOL_FALSE:
            self.index += 1
            return False
        raise ParsingError(f"expected bool token, got {token}")

    def read_str(self) -> str:
        token = self.data[self.index]
        if token.tag in (Tags.STR, Tags.LITERAL_STR):
            self.index += 1
            return cast(str, token.value)
        raise ParsingError(f"expected str token, got {token}")

    def read_ptr(self) -> AbstractEntity:
        token = self.data[self.index]
        if token.tag == Tags.POINTER:
            self.index += 1
            return cast(AbstractEntity, token.value)
        raise ParsingError(f"expected pointer token, got {token}")

    def read_transform(self) -> list[float]:
        # Transform matrix is stored as literal string like in the SAT format!
        # 4th column is not stored
        # Read only the matrix values which contain all information needed,
        # the additional data are only hints for the kernel how to process
        # the data (rotation, reflection, scaling, shearing).
        values = self.read_str().split(" ")
        return [float(v) for v in values[:12]]


class SabBuilder(AbstractBuilder):
    """Low level data structure to manage ACIS SAB data files."""

    def __init__(self) -> None:
        self.header = AcisHeader()
        self.bodies: list[SabEntity] = []
        self.entities: list[SabEntity] = []

    def dump_sab(self) -> bytes:
        """Returns the SAB representation of the ACIS file as bytes."""
        self.reorder_records()
        self.header.n_entities = len(self.bodies) + int(
            self.header.has_asm_header
        )
        self.header.n_records = 0  # is always 0
        self.header.flags = 12  # important for 21800 - meaning unknown
        data: list[bytes] = [self.header.dumpb()]
        encoder = Encoder()
        for record in build_sab_records(self.entities):
            encoder.write_record(record)
        data.extend(encoder.buffer)
        data.append(self.header.sab_end_marker())
        return b"".join(data)

    def set_entities(self, entities: list[SabEntity]) -> None:
        """Reset entities and bodies list. (internal API)"""
        self.bodies = [e for e in entities if e.name == "body"]
        self.entities = entities


class SabExporter(EntityExporter[SabEntity]):
    def make_record(self, entity: AcisEntity) -> SabEntity:
        record = SabEntity(entity.type, id=entity.id)
        record.attributes = NULL_PTR
        return record

    def make_data_exporter(self, record: SabEntity) -> DataExporter:
        return SabDataExporter(self, record.data)

    def dump_sab(self) -> bytes:
        builder = SabBuilder()
        builder.header = self.header
        builder.set_entities(self.export_records())
        return builder.dump_sab()


def build_entities(
    records: Iterable[SabRecord], version: int
) -> Iterator[SabEntity]:
    for record in records:
        assert record[0].tag == Tags.ENTITY_TYPE, "invalid entity-name tag"
        name = record[0].value  # 1. entity-name
        if name in const.DATA_END_MARKERS:
            yield SabEntity(name)
            return
        assert record[1].tag == Tags.POINTER, "invalid attribute pointer tag"
        attr = record[1].value  # 2. attribute record pointer
        id_ = -1
        if version >= 700:
            assert record[2].tag == Tags.INT, "invalid id tag"
            id_ = record[2].value  # 3. int id
            data = record[3:]
        else:
            data = record[2:]
        yield SabEntity(name, attr, id_, data)


def resolve_pointers(entities: list[SabEntity]) -> list[SabEntity]:
    def ptr(num: int) -> SabEntity:
        if num == -1:
            return NULL_PTR
        return entities[num]

    for entity in entities:
        entity.attributes = ptr(entity.attr_ptr)
        entity.attr_ptr = -1
        for index, token in enumerate(entity.data):
            if token.tag == Tags.POINTER:
                entity.data[index] = Token(token.tag, ptr(token.value))
    return entities


def parse_sab(data: Union[bytes, bytearray]) -> SabBuilder:
    """Returns the :class:`SabBuilder` for the ACIS :term:`SAB` file content
    given as string or list of strings.

    Raises:
        ParsingError: invalid or unsupported ACIS data structure

    """
    if not isinstance(data, (bytes, bytearray)):
        raise TypeError("expected bytes, bytearray")
    builder = SabBuilder()
    decoder = Decoder(data)
    builder.header = decoder.read_header()
    entities = list(
        build_entities(decoder.read_records(), builder.header.version)
    )
    builder.set_entities(resolve_pointers(entities))
    return builder


class SabDataExporter(DataExporter):
    def __init__(self, exporter: SabExporter, data: list[Token]):
        self.version = exporter.version
        self.exporter = exporter
        self.data = data

    def write_int(self, value: int, skip_sat=False) -> None:
        """There are sometimes additional int values in SAB files which are
        not present in SAT files, maybe reference counters e.g. vertex, coedge.
        """
        self.data.append(Token(Tags.INT, value))

    def write_double(self, value: float) -> None:
        self.data.append(Token(Tags.DOUBLE, value))

    def write_interval(self, value: float) -> None:
        if math.isinf(value):
            self.data.append(Token(Tags.BOOL_FALSE, False))  # infinite "I"
        else:
            self.data.append(Token(Tags.BOOL_TRUE, True))  # finite "F"
            self.write_double(value)

    def write_loc_vec3(self, value: Vec3) -> None:
        self.data.append(Token(Tags.LOCATION_VEC, value))

    def write_dir_vec3(self, value: Vec3) -> None:
        self.data.append(Token(Tags.DIRECTION_VEC, value))

    def write_bool(self, value: bool, true: str, false: str) -> None:
        if value:
            self.data.append(Token(Tags.BOOL_TRUE, True))
        else:
            self.data.append(Token(Tags.BOOL_FALSE, False))

    def write_str(self, value: str) -> None:
        self.data.append(Token(Tags.STR, value))

    def write_literal_str(self, value: str) -> None:
        self.data.append(Token(Tags.LITERAL_STR, value))

    def write_ptr(self, entity: AcisEntity) -> None:
        record = NULL_PTR
        if not entity.is_none:
            record = self.exporter.get_record(entity)
        self.data.append(Token(Tags.POINTER, record))

    def write_transform(self, data: list[str]) -> None:
        # The last space is important!
        self.write_literal_str(" ".join(data) + " ")


def encode_entity_type(name: str) -> list[Token]:
    if name == const.NULL_PTR_NAME:
        raise InvalidLinkStructure(
            f"invalid record type: {const.NULL_PTR_NAME}"
        )
    parts = name.split("-")
    tokens = [Token(Tags.ENTITY_TYPE_EX, part) for part in parts[:-1]]
    tokens.append(Token(Tags.ENTITY_TYPE, parts[-1]))
    return tokens


def encode_entity_ptr(entity: SabEntity, entities: list[SabEntity]) -> Token:
    if entity.is_null_ptr:
        return Token(Tags.POINTER, -1)
    try:
        return Token(Tags.POINTER, entities.index(entity))
    except ValueError:
        raise InvalidLinkStructure(
            f"entity {str(entity)} not in record storage"
        )


def build_sab_records(entities: list[SabEntity]) -> Iterator[SabRecord]:
    for entity in entities:
        record: list[Token] = []
        record.extend(encode_entity_type(entity.name))
        # 1. attribute record pointer
        record.append(encode_entity_ptr(entity.attributes, entities))
        # 2. int id
        record.append(Token(Tags.INT, entity.id))
        for token in entity.data:
            if token.tag == Tags.POINTER:
                record.append(encode_entity_ptr(token.value, entities))
            elif token.tag == Tags.ENTITY_TYPE:
                record.extend(encode_entity_type(token.value))
            else:
                record.append(token)
        yield record


END_OF_RECORD = bytes([Tags.RECORD_END.value])
TRUE_RECORD = bytes([Tags.BOOL_TRUE.value])
FALSE_RECORD = bytes([Tags.BOOL_FALSE.value])
SUBTYPE_START_RECORD = bytes([Tags.SUBTYPE_START.value])
SUBTYPE_END_RECORD = bytes([Tags.SUBTYPE_END.value])
SAB_ENCODING = "utf8"


class Encoder:
    def __init__(self) -> None:
        self.buffer: list[bytes] = []

    def write_record(self, record: SabRecord) -> None:
        for token in record:
            self.write_token(token)
        self.buffer.append(END_OF_RECORD)

    def write_token(self, token: Token) -> None:
        tag = token.tag
        if tag in (Tags.INT, Tags.POINTER, Tags.ENUM):
            assert isinstance(token.value, int)
            self.buffer.append(struct.pack("<Bi", tag, token.value))
        elif tag == Tags.DOUBLE:
            assert isinstance(token.value, float)
            self.buffer.append(struct.pack("<Bd", tag, token.value))
        elif tag == Tags.STR:
            assert isinstance(token.value, str)
            data = token.value.encode(encoding=SAB_ENCODING)
            self.buffer.append(struct.pack("<BB", tag, len(data)) + data)
        elif tag == Tags.LITERAL_STR:
            assert isinstance(token.value, str)
            data = token.value.encode(encoding=SAB_ENCODING)
            self.buffer.append(struct.pack("<Bi", tag, len(data)) + data)
        elif tag in (Tags.ENTITY_TYPE, Tags.ENTITY_TYPE_EX):
            assert isinstance(token.value, str)
            data = token.value.encode(encoding=SAB_ENCODING)
            self.buffer.append(struct.pack("<BB", tag, len(data)) + data)
        elif tag in (Tags.LOCATION_VEC, Tags.DIRECTION_VEC):
            v = token.value
            assert isinstance(v, Vec3)
            self.buffer.append(struct.pack("<B3d", tag, v.x, v.y, v.z))
        elif tag == Tags.BOOL_TRUE:
            self.buffer.append(TRUE_RECORD)
        elif tag == Tags.BOOL_FALSE:
            self.buffer.append(FALSE_RECORD)
        elif tag == Tags.SUBTYPE_START:
            self.buffer.append(SUBTYPE_START_RECORD)
        elif tag == Tags.SUBTYPE_END:
            self.buffer.append(SUBTYPE_END_RECORD)
        else:
            raise ValueError(f"invalid tag in token: {token}")



#  Copyright (c) 2020-2023, Manfred Moitzi
#  License: MIT License
# Test only basic features of Cython implementation,
# Full testing and compatibility check with Python implementation
# is located in test suite 605.
import pytest
import numpy as np

matrix44 = pytest.importorskip("ezdxf.acc.matrix44")
Matrix44: matrix44.Matrix44 = matrix44.Matrix44


def test_default_constructor():
    m = Matrix44()
    assert m[0, 0] == 1.0
    assert m[1, 1] == 1.0
    assert m[2, 2] == 1.0
    assert m[3, 3] == 1.0


def test_16_numbers_constructor():
    m = Matrix44(range(16))
    assert m[3, 3] == 15


def test_4_rows_constructor():
    m = Matrix44((0, 0, 0, 0), (1, 1, 1, 1), (2, 2, 2, 2), (3, 3, 3, 3))
    assert m[0, 0] == 0
    assert m[3, 3] == 3


def test_set_item():
    m = Matrix44()
    m[0, 0] = 17
    assert m[0, 0] == 17


def test_set_row_4_values():
    m = Matrix44()
    m.set_row(0, (2, 3, 4, 5))
    assert m.get_row(0) == (2, 3, 4, 5)


def test_set_row_1_value():
    m = Matrix44()
    m.set_row(1, (2,))
    assert m.get_row(1) == (2, 1, 0, 0)


def test_set_col_4_values():
    m = Matrix44()
    m.set_col(0, (2, 3, 4, 5))
    assert m.get_col(0) == (2, 3, 4, 5)


def test_set_col_1_value():
    m = Matrix44()
    m.set_col(1, (2,))
    assert m.get_col(1) == (2, 1, 0, 0)


def test_copy():
    m1 = Matrix44(range(16))
    m2 = m1.copy()
    assert m2.get_row(0) == (0, 1, 2, 3)
    m1.set_row(0, (20, 30, 40, 50))
    assert m1.get_row(0) == (20, 30, 40, 50)
    assert m2.get_row(0) == (0, 1, 2, 3)


def test_get_origin():
    m = Matrix44()
    m.set_row(3, (7, 8, 9))
    assert m.origin == (7, 8, 9)


def test_array_inplace_transformation():
    from ezdxf.math import Vec2

    m = matrix44.Matrix44.translate(1, 2, 0)
    points = [(0, 0), (1, 1), (2, 2)]
    control = m.fast_2d_transform(points)

    array = np.array(points, dtype=np.float64)
    m.transform_array_inplace(array, 2)
    assert all(Vec2(c).isclose(r) for c, r in zip(control, array))


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
"""
This module provides "nested Polygon" detection for multiple paths.

Terminology
-----------

exterior
    creates a filled area, has counter-clockwise (ccw) winding
    exterior := Path

hole
    creates an unfilled area, has clockwise winding (cw),
    hole := Polygon

polygon
    list of nested paths:
    polygon without a hole: [path]
    polygon with 1 hole: [path, [path]]
    polygon with 2 separated holes: [path, [path], [path]]
    polygon with 2 nested holes: [path, [path, [path]]]

    polygon := [exterior, *hole]

The result is a list of polygons:

1 polygon returns: [[ext-path]]
2 separated polygons returns: [[ext-path], [ext-path, [hole-path]]]

A hole is just another polygon, some render backends may require a distinct winding
order for nested paths like: ccw-cw-ccw-cw...

[Exterior-ccw,
    [Hole-Exterior-cw,
        [Sub-Hole-ccw],
        [Sub-Hole-ccw],
    ],
    [Hole-Exterior-cw],
    [Hole-Exterior-cw],
]

The implementation has to do some expensive tests, like check if a path is
inside of another path or if paths do overlap. A goal is to reduce this costs
by using proxy objects:

Bounding Box Proxy
------------------

This implementation uses the bounding box of the path as proxy object, this is very fast
but not accurate, but can handle most of the real world scenarios, in the assumption
that most HATCHES are created from non-overlapping boundary paths.
Overlap detection and resolving is not possible.

The input paths have to implement the SupportsBoundingBox protocol, which requires
only a method bbox() that returns a BoundingBox2d instance for the path.

Sort by Area
------------

It is not possible for a path to contain another path with a larger area.

"""
from __future__ import annotations
from typing import (
    Tuple,
    Optional,
    List,
    Iterable,
    Sequence,
    Iterator,
    TypeVar,
)
from typing_extensions import TypeAlias
from collections import namedtuple
from ezdxf.math import AbstractBoundingBox
from ezdxf.protocols import SupportsBoundingBox


__all__ = [
    "make_polygon_structure",
    "winding_deconstruction",
    "group_paths",
    "flatten_polygons",
]


T = TypeVar("T", bound=SupportsBoundingBox)

Polygon: TypeAlias = Tuple[T, Optional[List["Polygon"]]]
BoxStruct = namedtuple("BoxStruct", "bbox, path")


def make_polygon_structure(paths: Iterable[T]) -> list[Polygon]:
    """Returns a recursive polygon structure from iterable `paths`, uses 2D
    bounding boxes as fast detection objects.

    """

    # Implements fast bounding box construction and fast inside check.
    def area(item: BoxStruct) -> float:
        size = item.bbox.size
        return size.x * size.y

    def separate(
        exterior: AbstractBoundingBox, candidates: list[BoxStruct]
    ) -> tuple[list[BoxStruct], list[BoxStruct]]:
        holes: list[BoxStruct] = []
        outside: list[BoxStruct] = []
        for candidate in candidates:
            # Fast inside check:
            (holes if exterior.inside(candidate.bbox.center) else outside).append(
                candidate
            )
        return holes, outside

    def polygon_structure(outside: list[BoxStruct]) -> list[list]:
        polygons = []
        while outside:
            exterior = outside.pop()  # path with the largest area
            # Get holes inside of exterior and returns the remaining paths
            # outside of exterior:
            holes, outside = separate(exterior.bbox, outside)
            if holes:
                # build nested hole structure:
                # the largest hole could contain the smaller holes,
                # and so on ...
                holes = polygon_structure(holes)  # type: ignore
            polygons.append([exterior, *holes])
        return polygons

    def as_nested_paths(polygons) -> list:
        return [
            polygon.path if isinstance(polygon, BoxStruct) else as_nested_paths(polygon)
            for polygon in polygons
        ]

    boxed_paths = []
    for path in paths:
        bbox = path.bbox()
        if bbox.has_data:
            boxed_paths.append(BoxStruct(bbox, path))
    boxed_paths.sort(key=area)
    return as_nested_paths(polygon_structure(boxed_paths))


def winding_deconstruction(
    polygons: list[Polygon],
) -> tuple[list[T], list[T]]:
    """Flatten the nested polygon structure in a tuple of two lists,
    the first list contains the paths which should be counter-clockwise oriented
    and the second list contains the paths which should be clockwise oriented.

    The paths are not converted to this orientation.

    """

    def deconstruct(polygons_, level):
        for polygon in polygons_:
            if isinstance(polygon, Sequence):
                deconstruct(polygon, level + 1)
            else:
                # level 0 is the list of polygons
                # level 1 = ccw, 2 = cw, 3 = ccw, 4 = cw, ...
                (ccw_paths if (level % 2) else cw_paths).append(polygon)

    cw_paths: list[T] = []
    ccw_paths: list[T] = []
    deconstruct(polygons, 0)
    return ccw_paths, cw_paths


def flatten_polygons(polygons: Polygon) -> Iterator[T]:
    """Yield a flat representation of the given nested polygons."""
    for polygon in polygons:
        if isinstance(polygon, Sequence):
            yield from flatten_polygons(polygon)  # type: ignore
        else:
            yield polygon  # type: ignore  # T


def group_paths(paths: Iterable[T]) -> list[list[T]]:
    """Group separated paths and their inner holes as flat lists."""
    polygons = make_polygon_structure(paths)
    return [list(flatten_polygons(polygon)) for polygon in polygons]



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import math
from ezdxf.acis.api import mesh_from_body, load, body_from_mesh, vertices_from_body
from ezdxf.acis.mesh import PolyhedronFaceBuilder
from ezdxf.acis import entities, dbg
from ezdxf.render import forms
from ezdxf.math import BoundingBox, Vec3


class TestPrismToMesh:
    """The "prism" is a distorted cube with quadrilateral faces at the
    top and bottom and eight triangles as connecting faces in between.
    """

    @pytest.fixture(scope="class")
    def mesh(self, prism_sat):
        body = load(prism_sat)[0]
        return mesh_from_body(body)[0]

    def test_mesh_has_10_faces(self, mesh):
        assert len(mesh.faces) == 10

    def test_mesh_has_8_unique_vertices(self, mesh):
        assert len(mesh.vertices) == 8

    def test_all_faces_have_at_least_3_vertices(self, mesh):
        assert all(len(set(face)) >= 3 for face in mesh.faces)

    def test_any_face_has_at_least_4_vertices(self, mesh):
        assert any(len(set(face)) == 4 for face in mesh.faces)


class TestPolyhedronFaceBuilder:
    @pytest.fixture
    def builder(self):
        return PolyhedronFaceBuilder(forms.cube())

    def test_six_faces_will_be_created(self, builder):
        """There have to be 6 cube faces."""
        assert len(builder.acis_faces()) == 6

    def test_each_face_gets_it_own_plane(self, builder):
        """Each face has to have its own flat surface plane."""
        planes = {id(face.surface) for face in builder.acis_faces()}
        assert len(planes) == 6
        assert (
            all(face.surface.type == "plane-surface" for face in builder.acis_faces())
            is True
        )

    def test_each_face_has_a_single_loop(self, builder):
        """The PolyhedronFaceBuilder supports only a single loop for each face,
        so divided faces or holes in faces are not supported.
        """
        for face in builder.acis_faces():
            assert face.loop.next_loop.is_none is True

    def test_for_24_unique_coedges(self, builder):
        """Each face has its own coedges which represent the face boundary and
        each coedge references a 'real' edge entity
        """
        coedges = get_coedges(builder.acis_faces())
        assert len(coedges) == 24

    def test_coedge_references_parent_loop(self, builder):
        for coedge in get_coedges(builder.acis_faces()):
            assert coedge.loop.is_none is False

    def test_for_partner_coedges(self, builder):
        """Each coedge has to have a partner coedge in the adjacent face and
        because they are sharing the same 'real' edge they have to have an
        opposite sense.
        """
        for coedge in get_coedges(builder.acis_faces()):
            partner_coedge = coedge.partner_coedge
            assert partner_coedge.is_none is False
            assert coedge.sense is not partner_coedge.sense
            assert partner_coedge.partner_coedge is coedge, "is back linked"

    def test_for_12_unique_edges(self, builder):
        """There have to be 12 'real' cube edges."""
        edges = get_edges(builder.acis_faces())
        assert len(edges) == 12

    def test_edges_have_a_parent_coedge(self, builder):
        for edge in get_edges(builder.acis_faces()):
            assert edge.coedge.type == "coedge"

    def test_each_edges_has_a_unique_straight_line(self, builder):
        """Each edge has to have a straight lines as base curve."""
        edges = get_edges(builder.acis_faces())
        assert all(e.curve.type == "straight-curve" for e in edges) is True
        for e in edges:
            ray = cast(entities.StraightCurve, e.curve)
            assert math.isclose(ray.direction.magnitude, 1.0)
        assert len(edges) == 12

    def test_for_8_unique_vertices(self, builder):
        """Edges do share start- and end vertices which has a reference
        the 'real' point.
        """
        vertices = get_vertices(builder.acis_faces())
        assert len(vertices) == 8

    def test_for_8_unique_points(self, builder):
        """There have to be 8 'real' cube corner points."""
        points = get_points(builder.acis_faces())
        assert len(points) == 8

    def test_build_multiple_times_independent_faces(self, builder):
        """The two builds should not share any data."""
        faces1 = builder.acis_faces()
        faces2 = builder.acis_faces()
        assert set(faces1).isdisjoint(set(faces2))
        assert set(get_coedges(faces1)).isdisjoint(set(get_coedges(faces2)))
        assert set(get_edges(faces1)).isdisjoint(set(get_edges(faces2)))
        assert set(get_vertices(faces1)).isdisjoint(set(get_vertices(faces2)))
        assert set(get_points(faces1)).isdisjoint(set(get_points(faces2)))


def get_coedges(faces):
    coedges = set()
    for face in faces:
        for coedge in face.loop.coedges():
            coedges.add(coedge)
    return coedges


def get_edges(faces):
    edges = set()
    for face in faces:
        for coedge in face.loop.coedges():
            edges.add(coedge.edge)
    return edges


def get_vertices(faces):
    vertices = set()
    for face in faces:
        for coedge in face.loop.coedges():
            vertices.add(coedge.edge.start_vertex)
            vertices.add(coedge.edge.end_vertex)
    return vertices


def get_points(faces):
    points = set()
    for face in faces:
        for coedge in face.loop.coedges():
            points.add(coedge.edge.start_vertex.point)
            points.add(coedge.edge.end_vertex.point)
    return points


def test_rebuild_mesh_from_acis_body():
    """Convert a cube-mesh into an ACIS body and the ACIS body back to a
    MeshBuilder instance.
    """
    cube = forms.cube()
    body = body_from_mesh(cube)
    cube2 = mesh_from_body(body)[0]
    assert set(cube.vertices) == set(cube2.vertices)
    assert faces_set(cube) == faces_set(cube2)


def faces_set(mesh):
    return set([tuple(face) for face in mesh.faces_as_vertices()])


class TestTransformCenterOfMeshToOriginAtConversionToAcisBody:
    @pytest.fixture(scope="class")
    def body(self):
        cube = forms.cube(center=False)
        return body_from_mesh(cube)

    def test_transformation_matrix_for_translation(self, body):
        matrix = body.transform.matrix
        translation = Vec3(matrix.get_row(3)[:3])
        assert translation.isclose((0.5, 0.5, 0.5))

    def test_if_acis_points_are_centered_around_the_origin(self, body):
        debugger = dbg.AcisDebugger(body)
        vertices = [p.location for p in debugger.filter_type("point")]
        bbox = BoundingBox(vertices)
        assert bbox.extmin.isclose((-0.5, -0.5, -0.5))
        assert bbox.extmax.isclose((0.5, 0.5, 0.5))


def test_non_manifold_detection():
    cube = forms.cube(center=False)
    # duplicate last face
    cube.faces.append(cube.faces[-1])
    body = body_from_mesh(cube)
    debugger = dbg.AcisDebugger(body)
    assert debugger.is_manifold() is False


def test_body_from_menger_sponge_is_manifold():
    from ezdxf.addons import MengerSponge

    sponge = MengerSponge().mesh()
    body = body_from_mesh(sponge)
    debugger = dbg.AcisDebugger(body)
    assert debugger.is_manifold() is True


def test_vertices_from_body():
    cube = forms.cube()
    body = body_from_mesh(cube)
    vertices = vertices_from_body(body)
    extents = BoundingBox(vertices)

    assert extents.extmin.isclose((-0.5, -0.5, -0.5))
    assert extents.extmax.isclose((0.5, 0.5, 0.5))
    assert len(vertices) == 6 * 4 * 2  # 6 faces x 4 edges x 2 vertices


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.math import Vec2
from ezdxf.path import Path
from ezdxf.npshapes import NumpyPath2d
from ezdxf.addons.drawing import hpgl2, layout
from ezdxf.addons.drawing.properties import BackendProperties


def test_empty_page():
    backend = hpgl2.PlotterBackend()
    backend.draw_point(Vec2(0, 0), BackendProperties())
    assert backend.get_bytes(layout.Page(0, 0)) == b""


class TestPlotterBackend:
    @pytest.fixture(scope="class")
    def properties(self):
        return BackendProperties(color="#ff0000", lineweight=0.25)

    @pytest.fixture(scope="class")
    def result1(self, properties):
        backend = hpgl2.PlotterBackend()
        backend.draw_line(Vec2(0, 0), Vec2(100, 100), properties)
        return backend.compatible(layout.Page(0, 0))

    def test_enter_hpgl2_mode_prefix(self,result1):
        assert result1.startswith(b"%0B;")

    def test_init_commands_exist(self, result1):
        assert b"IN;" in result1
        assert b"BP;" in result1
        assert b"PS4000,4000;" in result1  # 40 plu = 1mm

    def test_pen_color_setup_exist(self, result1):
        assert b"PC1,255,0,0;" in result1

    def test_set_pen_command_exist(self, result1):
        assert b"SP1;" in result1

    def test_pen_width_command_exist(self, result1):
        assert b"PW0.25;" in result1

    def test_lines_are_7_bit_int_base_32_polyline_encoded(self, result1):
        assert b"PE7<=__?Yf?Yf;" in result1

    def test_curves_are_ascii_encoded_relative_bezier_curves(self, properties):
        path = Path((0, 0))
        path.curve4_to((100, 100), (25, 0), (75, 100))
        backend = hpgl2.PlotterBackend()
        backend.draw_path(NumpyPath2d(path), properties)
        result = backend.high_quality(layout.Page(0, 0))
        assert b"PU;PA0,0;PD;BR1000,0,3000,4000,4000,4000;PU;" in result


if __name__ == '__main__':
    pytest.main([__file__])




from typing import List
import pytest
import ezdxf
import math
from math import isclose
from ezdxf.math import Vec3, global_bspline_interpolation, close_vectors, cad_fit_point_interpolation
from ezdxf.math.parametrize import (
    uniform_t_vector,
    distance_t_vector,
    centripetal_t_vector,
    arc_t_vector,
    arc_distances,
    estimate_tangents,
)
from ezdxf.math.bspline import (
    knots_from_parametrization,
    required_knot_values,
    averaged_knots_unconstrained,
    natural_knots_constrained,
    averaged_knots_constrained,
    natural_knots_unconstrained,
    double_knots,
    create_t_vector,
)

POINTS1 = Vec3.list([(1, 1), (2, 4), (4, 1), (7, 6)])
POINTS2 = Vec3.list([(1, 1), (2, 4), (4, 1), (7, 6), (5, 8), (3, 3), (1, 7)])


@pytest.fixture(params=[POINTS1, POINTS2])
def fit_points(request):
    return request.param


def test_uniform_t_array(fit_points):
    t_vector = list(uniform_t_vector(len(fit_points)))
    assert len(t_vector) == len(fit_points)
    assert t_vector[0] == 0.0
    assert t_vector[-1] == 1.0
    for t1, t2 in zip(t_vector, t_vector[1:]):
        assert t1 <= t2


def test_chord_length_t_array(fit_points):
    t_vector = list(distance_t_vector(fit_points))
    assert len(t_vector) == len(fit_points)
    assert t_vector[0] == 0.0
    assert t_vector[-1] == pytest.approx(1.0)
    for t1, t2 in zip(t_vector, t_vector[1:]):
        assert t1 <= t2


def test_centripetal_length_t_array(fit_points):
    t_vector = list(centripetal_t_vector(fit_points))
    assert len(t_vector) == len(fit_points)
    assert t_vector[0] == 0.0
    assert t_vector[-1] == pytest.approx(1.0)
    for t1, t2 in zip(t_vector, t_vector[1:]):
        assert t1 <= t2


def test_arc_distances():
    p = Vec3.list([(0, 0), (2, 2), (4, 0), (6, -2), (8, 0)])
    # p[1]..p[3] are a straight line, radius calculation fails and
    # a straight line from p[1] to p[2] is used as replacement
    # for the second arc
    radius = 2.0
    arc_length = math.pi * 0.5 * radius
    diagonal = math.sqrt(2.0) * radius
    distances = list(arc_distances(p))
    assert len(distances) == 4
    assert isclose(distances[0], arc_length)
    assert isclose(distances[1], diagonal)  # replacement for arc
    assert isclose(distances[2], arc_length)
    assert isclose(distances[3], arc_length)


def test_arc_length_t_array(fit_points):
    t_vector = list(arc_t_vector(fit_points))
    assert len(t_vector) == len(fit_points)
    assert t_vector[0] == 0.0
    assert t_vector[-1] == pytest.approx(1.0)
    for t1, t2 in zip(t_vector, t_vector[1:]):
        assert t1 <= t2


def test_invalid_order_count_combination():
    count = 4
    order = 5
    with pytest.raises(ezdxf.DXFValueError):
        required_knot_values(count, order)
    with pytest.raises(ezdxf.DXFValueError):
        list(knots_from_parametrization(n=count - 1, p=order - 1, t=[]))


def check_knots(count: int, order: int, knots: List[float]):
    assert len(knots) == required_knot_values(count, order)
    assert len(set(knots[:order])) == 1, "first order elements have to be equal"
    assert len(set(knots[-order:])) == 1, "last order elements have to be equal"
    for k1, k2 in zip(knots, knots[1:]):
        assert k1 <= k2


@pytest.mark.parametrize("p", (2, 3, 4))
@pytest.mark.parametrize("method", ("average", "natural"))
def test_knot_generation(p, method):
    fit_points = Vec3.list(
        [
            (0, 0),
            (0, 10),
            (10, 10),
            (20, 10),
            (20, 0),
            (30, 0),
            (30, 10),
            (40, 10),
            (40, 0),
        ]
    )
    count = len(fit_points)
    n = count - 1
    order = p + 1
    t_vector = distance_t_vector(fit_points)
    knots = list(knots_from_parametrization(n, p, t_vector, method))
    check_knots(n + 1, p + 1, knots)


@pytest.fixture
def fit_points_2():
    return Vec3.list(
        [
            (0, 0),
            (0, 10),
            (10, 10),
            (20, 10),
            (20, 0),
            (30, 0),
            (30, 10),
            (40, 10),
            (40, 0),
        ]
    )


@pytest.mark.parametrize("p", (2, 3, 4, 5))
def test_unconstrained_averaged_knots(p, fit_points_2):
    t_vector = list(distance_t_vector(fit_points_2))
    n = len(fit_points_2) - 1

    knots = averaged_knots_unconstrained(n, p, t_vector)
    check_knots(n + 1, p + 1, knots)


@pytest.mark.parametrize("p", (2, 3, 4, 5))
def test_constrained_averaged_knots(p, fit_points_2):
    t_vector = list(distance_t_vector(fit_points_2))
    n = len(fit_points_2) - 1

    # add 2 knots for tangents
    knots = averaged_knots_constrained(n + 2, p, t_vector)
    check_knots(n + 3, p + 1, knots)


@pytest.mark.parametrize("p", (2, 3, 4, 5))
def test_unconstrained_natural_knots(p, fit_points_2):
    t_vector = list(distance_t_vector(fit_points_2))
    n = len(fit_points_2) - 1

    # add 2 knots for tangents
    knots = natural_knots_unconstrained(n, p, t_vector)
    check_knots(n + 1, p + 1, knots)


@pytest.mark.parametrize("p", (2, 3, 4, 5))
def test_constrained_natural_knots(p, fit_points_2):
    t_vector = list(distance_t_vector(fit_points_2))
    n = len(fit_points_2) - 1

    # add 2 knots for tangents
    knots = natural_knots_constrained(n + 2, p, t_vector)
    check_knots(n + 3, p + 1, knots)


@pytest.mark.parametrize("p", (2, 3, 4, 5))
def test_double_knots(p, fit_points_2):
    t_vector = list(distance_t_vector(fit_points_2))
    n = len(fit_points_2) - 1

    # create knots for each control point and 1st derivative
    knots = double_knots(n, p, t_vector)
    check_knots((n + 1) * 2, p + 1, knots)


def test_bspline_interpolation(fit_points):
    spline = global_bspline_interpolation(fit_points, degree=3, method="chord")
    assert len(spline.control_points) == len(fit_points)

    t_array = list(create_t_vector(fit_points, "chord"))
    assert t_array[0] == 0.0
    assert t_array[-1] == pytest.approx(1.0)
    assert len(t_array) == len(fit_points)

    t_points = [spline.point(t) for t in t_array]
    assert close_vectors(t_points, fit_points)


@pytest.mark.parametrize("method", ["distance", "centripetal", "arc"])
def test_create_t_vectors_for_identical_points(method):
    assert len(list(create_t_vector(Vec3.list([(0, 0), (0, 0)]), method))) == 0


def test_bspline_interpolation_first_derivatives(fit_points):
    tangents = estimate_tangents(fit_points)
    spline = global_bspline_interpolation(
        fit_points, degree=3, tangents=tangents
    )
    assert len(spline.control_points) == 2 * len(fit_points)


expected = [
    (0.0, 0.0),
    (0.010310831479728222, 0.32375901937484741),
    (0.030277278274297714, 0.6140977144241333),
    (0.059526983648538589, 0.87245100736618042),
    (0.097687594592571259, 1.1002539396286011),
    (0.14438676834106445, 1.2989413738250732),
    (0.19925214350223541, 1.469948410987854),
    (0.26191136240959167, 1.6147100925445557),
    (0.33199205994606018, 1.7346612215042114),
    (0.40912193059921265, 1.8312369585037231),
    (0.49292856454849243, 1.9058722257614136),
    (0.58303964138031006, 1.9600019454956055),
    (0.67908281087875366, 1.9950611591339111),
    (0.78068572282791138, 2.0124847888946533),
    (0.88747602701187134, 2.0137078762054443),
    (0.9990813136100769, 2.0001654624938965),
    (1.1151292324066162, 1.973292350769043),
    (1.2352476119995117, 1.9345237016677856),
    (1.3590638637542725, 1.8852944374084473),
    (1.4862056970596313, 1.8270395994186401),
    (1.6163008213043213, 1.761194109916687),
    (1.7489768266677856, 1.6891928911209106),
    (1.8838614225387573, 1.6124709844589233),
    (2.0205821990966797, 1.532463550567627),
    (2.1587669849395752, 1.4506052732467651),
    (2.2980430126190186, 1.3683313131332397),
    (2.4380383491516113, 1.2870765924453735),
    (2.5783803462982178, 1.2082761526107788),
    (2.7186968326568604, 1.1333650350570679),
    (2.8586153984069824, 1.0637780427932739),
    (2.9977638721466064, 1.0009502172470093),
    (3.1357693672180176, 0.94631654024124146),
    (3.2722601890563965, 0.90131211280822754),
    (3.4068636894226074, 0.86737185716629028),
    (3.5392072200775146, 0.84593069553375244),
    (3.6689188480377197, 0.83842372894287109),
    (3.795626163482666, 0.84628582000732422),
    (3.9189567565917969, 0.87095201015472412),
    (4.0385379791259766, 0.91385728120803833),
    (4.1539978981018066, 0.97643661499023438),
    (4.2649641036987305, 1.0601249933242798),
    (4.3710641860961914, 1.1663573980331421),
    (4.4719257354736328, 1.2965688705444336),
    (4.567176342010498, 1.4521942138671875),
    (4.6564435958862305, 1.6346687078475952),
    (4.7393555641174316, 1.8454270362854004),
    (4.8155393600463867, 2.0859043598175049),
    (4.8846230506896973, 2.3575356006622314),
    (4.9462337493896484, 2.6617558002471924),
    (5.0, 3.0),
]


def test_check_values():
    test_points = [(0.0, 0.0), (1.0, 2.0), (3.0, 1.0), (5.0, 3.0)]
    spline = global_bspline_interpolation(
        test_points, degree=3, method="distance"
    )
    result = list(spline.approximate(49))
    assert len(result) == 50
    for p1, p2 in zip(result, expected):
        assert isclose(p1[0], p2[0], abs_tol=1e-6)
        assert isclose(p1[1], p2[1], abs_tol=1e-6)


def test_cad_fit_point_interpolation_for_2_points():
    points = Vec3.list([(0, 0), (0, 10)])
    control_points, knots = cad_fit_point_interpolation(points)
    assert control_points[0].isclose(points[0])
    assert control_points[-1].isclose(points[-1])
    assert len(control_points) == 4
    assert len(knots) == required_knot_values(4, 4) == 8


def test_cad_fit_point_interpolation_for_5_points():
    points = Vec3.list([(0, 0), (0, 10), (10, 10), (20, 10), (20, 0)])
    control_points, knots = cad_fit_point_interpolation(points)
    assert len(control_points) == 7
    assert len(knots) == required_knot_values(7, 4) == 11

    # Checked visually by BricsCAD 2022 and TrueView 2022:
    # See function check_visually_fit_points_to_cad_cv() in script
    # exploration/spline_end_tangent_estimation.py
    assert control_points[0].isclose(points[0])
    assert control_points[1].isclose(Vec3(-0.8333333333333334, 4.285714285714286, -0.0))
    assert control_points[2].isclose(Vec3(-2.5, 12.857142857142858, 0.0))
    assert control_points[3].isclose(Vec3(10.0, 8.571428571428573, 0.0))
    assert control_points[4].isclose(Vec3(22.5, 12.857142857142858, 0.0))
    assert control_points[5].isclose(Vec3(20.833333333333332, 4.2857142857142865, -0.0))
    assert control_points[6].isclose(points[-1])





#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import is_face_normal_pointing_outwards
from ezdxf.render import forms


def test_cube_with_ccw_vertex_orientation():
    mesh = forms.cube()
    faces = list(mesh.faces_as_vertices())
    for index, face in enumerate(faces):
        assert (
            is_face_normal_pointing_outwards(faces, face) is True
        ), f"index = {index}"


def test_cube_with_clockwise_vertex_orientation():
    mesh = forms.cube()
    mesh.flip_normals()
    faces = list(mesh.faces_as_vertices())
    for index, face in enumerate(faces):
        assert (
            is_face_normal_pointing_outwards(faces, face) is False
        ), f"index = {index}"


COUNT = 11


def test_torus_with_ccw_vertex_orientation():
    mesh = forms.torus(minor_count=COUNT)
    faces = list(mesh.faces_as_vertices())
    for index, face in enumerate(faces):
        assert (
            is_face_normal_pointing_outwards(faces, face) is True
        ), f"index = {index}"


def test_torus_with_clockwise_vertex_orientation():
    mesh = forms.torus(minor_count=COUNT)
    mesh.flip_normals()
    faces = list(mesh.faces_as_vertices())
    for index, face in enumerate(faces):
        assert (
            is_face_normal_pointing_outwards(faces, face) is False
        ), f"index = {index}"


def test_flipped_cone():
    cone = forms.cone(3)
    cone.flip_normals()
    faces = list(cone.faces_as_vertices())
    for index, face in enumerate(faces):
        assert (
            is_face_normal_pointing_outwards(faces, face) is False
        ), f"index = {index}"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
from typing import cast
import pytest
import ezdxf
from ezdxf.entities.idbuffer import FieldList
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

FIELDLIST = """0
FIELDLIST
5
0
102
{ACAD_REACTORS
330
0
102
}
330
0
100
AcDbIdSet
90
12
100
AcDbFieldList
"""


@pytest.fixture
def entity():
    return FieldList.from_text(FIELDLIST)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "FIELDLIST" in ENTITY_CLASSES


def test_default_init():
    entity = FieldList()
    assert entity.dxftype() == "FIELDLIST"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = FieldList.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.flags == 0
    assert len(entity.handles) == 0


def test_load_from_text(entity):
    assert entity.dxf.flags == 12
    assert len(entity.handles) == 0


def test_write_dxf():
    entity = FieldList.from_text(FIELDLIST)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(FIELDLIST)
    assert result == expected


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2007")


def test_generic_field_list(doc):
    field_list = doc.objects.new_entity("FIELDLIST", {})
    assert field_list.dxftype() == "FIELDLIST"
    assert len(field_list.handles) == 0


def test_set_get_field_list(doc):
    field_list = doc.objects.new_entity("FIELDLIST", {})
    assert field_list.dxftype() == "FIELDLIST"
    field_list.handles = ["FF", "EE", "DD"]
    handles = field_list.handles
    assert len(handles) == 3
    assert handles == ["FF", "EE", "DD"]

    handles.append("FFFF")
    assert handles[-1] == "FFFF"


def test_dxf_tags(doc):
    buffer = cast(FieldList, doc.objects.new_entity("FIELDLIST", {}))
    buffer.handles = ["FF", "EE", "DD", "CC"]
    tags = TagCollector.dxftags(buffer)

    assert len(tags) == 10
    assert tags[5] == (330, "FF")
    assert tags[8] == (330, "CC")
    assert tags[-1] == (100, "AcDbFieldList")


def test_clone(doc):
    buffer = cast(FieldList, doc.objects.new_entity("FIELDLIST", {}))
    buffer.handles = ["FF", "EE", "DD", "CC"]
    buffer2 = cast(FieldList, buffer.copy())
    buffer2.handles[-1] = "ABCD"
    assert buffer.handles[:-1] == buffer2.handles[:-1]
    assert buffer.handles[-1] != buffer2.handles[-1]



#  Copyright (c) 2020-2024, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import copy
import ezdxf
from ezdxf.math import Vec3, Matrix44
from ezdxf.entities import (
    factory,
    Hatch,
    MPolygon,
    LWPolyline,
    BoundaryPathType,
)
from ezdxf.entities.polygon import DXFPolygon
from ezdxf.addons import geo
from ezdxf.render.forms import square, translate

EXTERIOR = [(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)]
EXTERIOR_Z = [(0, 0, 1), (10, 0, 1), (10, 10, 1), (0, 10, 1), (0, 0, 1)]
HOLE1 = [(1, 1), (1, 2), (2, 2), (2, 1), (1, 1)]
HOLE2 = [(3, 3), (3, 4), (4, 4), (4, 3), (3, 3)]

POINT = {"type": "Point", "coordinates": (0, 0)}
POINT_Z = {"type": "Point", "coordinates": (0, 0, 1)}
LINE_STRING = {"type": "LineString", "coordinates": EXTERIOR}
LINE_STRING_Z = {"type": "LineString", "coordinates": EXTERIOR_Z}
POLYGON_0 = {"type": "Polygon", "coordinates": [EXTERIOR]}
POLYGON_0_Z = {"type": "Polygon", "coordinates": [EXTERIOR_Z]}
POLYGON_1 = {"type": "Polygon", "coordinates": [EXTERIOR, HOLE1]}
POLYGON_2 = {"type": "Polygon", "coordinates": [EXTERIOR, HOLE1, HOLE2]}
MULTI_POINT = {
    "type": "MultiPoint",
    "coordinates": EXTERIOR,
}
MULTI_POINT_Z = {
    "type": "MultiPoint",
    "coordinates": EXTERIOR_Z,
}
MULTI_LINE_STRING = {
    "type": "MultiLineString",
    "coordinates": [EXTERIOR, HOLE1, HOLE2],
}
MULTI_LINE_STRING_Z = {
    "type": "MultiLineString",
    "coordinates": [EXTERIOR_Z],
}
MULTI_POLYGON = {
    "type": "MultiPolygon",
    "coordinates": [
        [EXTERIOR],
        [EXTERIOR, HOLE1],
        [EXTERIOR, HOLE1, HOLE2],
    ],
}
MULTI_POLYGON_Z = {
    "type": "MultiPolygon",
    "coordinates": [[EXTERIOR_Z]],
}

GEOMETRY_COLLECTION = {
    "type": "GeometryCollection",
    "geometries": [
        POINT,
        LINE_STRING,
        POLYGON_0,
    ],
}
FEATURE_1 = {
    "type": "Feature",
    "prop0": "property",
    "geometry": LINE_STRING,
}
FEATURE_2 = {
    "type": "Feature",
    "prop0": "property",
    "geometry": LINE_STRING,
}

FEATURE_COLLECTION = {
    "type": "FeatureCollection",
    "features": [FEATURE_1, FEATURE_2],
}

FEATURE_PROPERTIES = {
    "type": "Feature",
    "properties": {"layer": "GeoJSON"},
    "geometry": LINE_STRING,
}

FEATURE_PROPERTIES_GC = {
    "type": "Feature",
    "properties": {"layer": "GeoJSON"},
    "geometry": GEOMETRY_COLLECTION,
}


@pytest.mark.parametrize(
    "points",
    [
        [],
        [(0, 0)],
        [(0, 0), (1, 0)],
    ],
)
def test_polygon_mapping_vertex_count_error(points):
    with pytest.raises(ValueError):
        geo.polygon_mapping(Vec3.list(points), [])


def test_map_dxf_point():
    point = factory.new("POINT", dxfattribs={"location": (0, 0)})
    assert geo.mapping(point) == {"type": "Point", "coordinates": (0, 0)}


def test_map_dxf_line():
    point = factory.new("LINE", dxfattribs={"start": (0, 0), "end": (1, 0)})
    assert geo.mapping(point) == {
        "type": "LineString",
        "coordinates": [(0, 0), (1, 0)],
    }


def test_map_polyline():
    pline = cast("Polyline", factory.new("POLYLINE"))
    pline.append_vertices([(0, 0), (1, 0), (1, 1)])
    pline.close()
    assert geo.mapping(pline) == {
        "type": "Polygon",
        "coordinates": ([(0, 0), (1, 0), (1, 1), (0, 0)], []),
    }
    assert geo.mapping(pline, force_line_string=True) == {
        "type": "LineString",
        "coordinates": [(0, 0), (1, 0), (1, 1), (0, 0)],
    }


@pytest.mark.parametrize("dxftype", ["HATCH", "MPOLYGON"])
def test_map_dxf_polygon(dxftype: str):
    hatch = cast(
        Hatch,
        factory.new(
            dxftype,
            dxfattribs={
                "hatch_style": 0,
            },
        ),
    )
    hatch.paths.add_polyline_path(EXTERIOR, flags=1)  # EXTERNAL
    hatch.paths.add_polyline_path(HOLE1, flags=0)  # DEFAULT
    hatch.paths.add_polyline_path(HOLE2, flags=0)  # DEFAULT
    m = geo.mapping(hatch)
    assert m["type"] == "Polygon"
    exterior, holes = m["coordinates"]
    assert len(exterior) == 5  # vertices
    assert len(holes) == 2
    assert len(holes[0]) == 5  # vertices
    assert len(holes[1]) == 5  # vertices


def test_map_circle():
    circle = factory.new("CIRCLE")
    m = geo.mapping(circle)
    assert m["type"] == "Polygon"
    assert len(m["coordinates"][0]) == 8
    m = geo.mapping(circle, force_line_string=True)
    assert m["type"] == "LineString"


def test_map_arc():
    arc = factory.new(
        "ARC",
        dxfattribs={
            "start_angle": 0,
            "end_angle": 90,
        },
    )
    m = geo.mapping(arc)
    assert m["type"] == "LineString"
    assert len(m["coordinates"][0]) > 1


def test_arc_geo_proxy_wcs_to_crs():
    arc = factory.new(
        "ARC",
        dxfattribs={
            "start_angle": 0,
            "end_angle": 90,
        },
    )
    geo_proxy = geo.proxy(arc)
    geo_proxy.wcs_to_crs(Matrix44())
    assert len(geo_proxy.__geo_interface__["coordinates"][0]) > 1


@pytest.mark.parametrize(
    "entity",
    [
        {"type": "Point", "coordinates": (0, 0)},
        {"type": "LineString", "coordinates": [(0, 0), (1, 0)]},
        {"type": "MultiPoint", "coordinates": [(0, 0), (1, 0)]},
        {
            "type": "MultiLineString",
            "coordinates": [[(0, 0), (1, 0)], [(0, 0), (1, 0)]],
        },
        {
            "type": "Feature",
            "geometry": {"type": "Point", "coordinates": (0, 0)},
        },
        {
            "type": "GeometryCollection",
            "geometries": [{"type": "Point", "coordinates": (0, 0)}],
        },
        {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": (0, 0)},
                }
            ],
        },
    ],
)
def test_parse_types(entity):
    # Parser does basic structure validation and converts all coordinates into
    # Vec3 objects.
    assert geo.parse(entity) == entity


def test_parsing_type_error():
    with pytest.raises(TypeError):
        geo.parse({"type": "XXX"})


@pytest.mark.parametrize(
    "entity",
    [
        {"type": "Point"},  # no coordinates key
        {"type": "Point", "coordinates": None},  # no coordinates
        {"type": "Feature"},  # no geometry key
        {"type": "GeometryCollection"},  # no geometries key
        {"type": "FeatureCollection"},  # no features key
    ],
)
def test_parsing_value_error(entity):
    with pytest.raises(ValueError):
        geo.parse(entity)


def test_parse_polygon_without_holes():
    polygon = geo.parse(POLYGON_0)
    assert polygon["coordinates"] == (EXTERIOR, [])


def test_parse_polygon_1_hole():
    polygon = geo.parse(POLYGON_1)
    assert polygon["coordinates"] == (EXTERIOR, [HOLE1])


def test_parse_polygon_2_holes():
    polygon = geo.parse(POLYGON_2)
    assert polygon["coordinates"] == (EXTERIOR, [HOLE1, HOLE2])


def test_parse_geometry_collection():
    geometry_collection = geo.parse(GEOMETRY_COLLECTION)
    assert len(geometry_collection["geometries"]) == 3


def test_parse_feature():
    feature = geo.parse(FEATURE_1)
    assert feature["geometry"] == LINE_STRING


def test_feature_with_geometry_collection():
    feature = geo.parse(FEATURE_PROPERTIES_GC)
    geometry_collection = feature["geometry"]
    assert geometry_collection["type"] == "GeometryCollection"
    geometries = geometry_collection["geometries"]
    assert isinstance(geometries, list)


def test_parse_feature_collection():
    feature_collection = geo.parse(FEATURE_COLLECTION)
    assert len(feature_collection["features"]) == 2


def test_iter_feature_with_geometry_collection():
    gp = geo.GeoProxy(FEATURE_PROPERTIES_GC)
    entities = list(gp)
    assert len(entities) == 3


@pytest.mark.parametrize(
    "entity",
    [
        POINT,
        POINT_Z,
        LINE_STRING,
        LINE_STRING_Z,
        POLYGON_0,
        POLYGON_0_Z,
        POLYGON_1,
        POLYGON_2,
        GEOMETRY_COLLECTION,
        FEATURE_1,
        FEATURE_COLLECTION,
        MULTI_POINT,
        MULTI_POINT_Z,
        MULTI_LINE_STRING,
        MULTI_LINE_STRING_Z,
        MULTI_POLYGON,
        MULTI_POLYGON_Z,
    ],
)
def test_geo_interface_builder(entity):
    assert geo.GeoProxy.parse(entity).__geo_interface__ == entity


@pytest.mark.parametrize(
    "entity, expected_coords",
    [
        (
            {"type": "MultiPoint", "coordinates": [(0, 0), (0, 1, 1)]},
            [(0, 0, 0), (0, 1, 1)],
        ),
        (
            {"type": "LineString", "coordinates": [(0, 0), (0, 1, 1)]},
            [(0, 0, 0), (0, 1, 1)],
        ),
        (
            {"type": "MultiLineString", "coordinates": [[(0, 0), (0, 1, 1)]]},
            [[(0, 0, 0), (0, 1, 1)]],
        ),
        (
            {"type": "Polygon", "coordinates": [[(0, 0), (0, 1, 1)]]},
            [[(0, 0, 0), (0, 1, 1)]],
        ),
        (
            {"type": "MultiPolygon", "coordinates": [[[(0, 0), (0, 1, 1)]]]},
            [[[(0, 0, 0), (0, 1, 1)]]],
        ),
    ],
)
def test_geo_interface_builder_given_heterogeneous_z_returns_3d_geometry(
    entity, expected_coords
) -> None:
    """This test verifies that if a geometry contains at least one non-zero Z component,
    it is processed as a 3D geometry. The Z components that were not specified are
    set to 0.
    """
    assert (
        geo.GeoProxy.parse(entity).__geo_interface__["coordinates"] == expected_coords
    )


def test_point_to_dxf_entity():
    point = list(geo.dxf_entities(POINT))[0]
    assert point.dxftype() == "POINT"
    assert point.dxf.location == (0, 0)


def test_line_string_to_dxf_entity():
    res = cast(LWPolyline, list(geo.dxf_entities(LINE_STRING))[0])
    assert res.dxftype() == "LWPOLYLINE"
    assert list(res.vertices()) == Vec3.list(EXTERIOR)


@pytest.mark.parametrize("dxftype, polygon", [("HATCH", 1), ("MPOLYGON", 4)])
def test_polygon_without_holes_to_dxf_polygon(dxftype, polygon):
    entity = cast(DXFPolygon, list(geo.dxf_entities(POLYGON_0, polygon=polygon))[0])
    assert entity.dxftype() == dxftype
    assert len(entity.paths) == 1
    p = entity.paths[0]
    assert p.type == BoundaryPathType.POLYLINE
    assert p.vertices == Vec3.list(EXTERIOR)


@pytest.mark.parametrize("dxftype, polygon", [("HATCH", 1), ("MPOLYGON", 4)])
def test_polygon_with_holes_to_dxf_polygon(dxftype, polygon):
    entity = cast(DXFPolygon, list(geo.dxf_entities(POLYGON_2, polygon=polygon))[0])
    assert entity.dxftype() == dxftype
    assert len(entity.paths) == 3
    p = entity.paths[1]
    assert p.type == BoundaryPathType.POLYLINE
    assert p.vertices == Vec3.list(HOLE1)
    p = entity.paths[2]
    assert p.type == BoundaryPathType.POLYLINE
    assert p.vertices == Vec3.list(HOLE2)


def test_mpolygon_supports_fill_and_border_color():
    attribs = {"color": 2, "fill_color": 3}
    mpolygon = cast(
        MPolygon,
        list(geo.dxf_entities(POLYGON_2, polygon=4, dxfattribs=attribs))[0],
    )
    assert mpolygon.dxftype() == "MPOLYGON"
    assert mpolygon.dxf.color == 2
    assert mpolygon.dxf.fill_color == 3


def test_geometry_collection_to_dxf_entities():
    collection = list(geo.dxf_entities(GEOMETRY_COLLECTION))
    assert len(collection) == 3


def test_feature_to_dxf_entities():
    entities = list(geo.dxf_entities(FEATURE_1))
    assert entities[0].dxftype() == "LWPOLYLINE"


def test_feature_collection_to_dxf_entities():
    collection = list(geo.dxf_entities(FEATURE_COLLECTION))
    assert len(collection) == 2
    assert collection[0].dxftype() == "LWPOLYLINE"


def test_dxf_entities_post_process_properties():
    entities = list(
        geo.dxf_entities(FEATURE_PROPERTIES, post_process=geo.assign_layers)
    )
    polyline = entities[0]
    assert polyline.dxf.layer == "GeoJSON"


def test_dxf_entities_from_geometry_collection_post_process_properties():
    entities = list(
        geo.dxf_entities(FEATURE_PROPERTIES_GC, post_process=geo.assign_layers)
    )
    assert len(entities) == 3
    for e in entities:
        assert e.dxf.layer == "GeoJSON"


@pytest.mark.parametrize(
    "deg, coords",
    [
        [(15, 47), (1669792.36, 5910809.62)],
        [(-15, 47), (-1669792.36, 5910809.62)],
        [(15, -47), (1669792.36, -5910809.62)],
        [(-15, -47), (-1669792.36, -5910809.62)],
        [(0, 0), (0, 0)],
    ],
)
def test_common_WGS84_projection(deg, coords):
    projected = geo.wgs84_4326_to_3395(Vec3(deg))
    assert projected.round(2).isclose(coords)
    # inverse projection
    assert geo.wgs84_3395_to_4326(projected).isclose(deg)


def validate(p: geo.GeoProxy):
    return p.geotype == "Point"


@pytest.mark.parametrize(
    "entity,type_",
    [
        [POINT, "Point"],
        [LINE_STRING, None],
        [POLYGON_0, None],
        [MULTI_POINT, "MultiPoint"],
        [MULTI_LINE_STRING, None],
        [MULTI_POLYGON, None],
        [FEATURE_1, None],
    ],
)
def test_filter_function_single_entity(entity, type_):
    p = geo.GeoProxy(copy.deepcopy(entity))
    p.filter(validate)
    assert p.geotype == type_


def test_filter_function_geometrie_collection():
    p = geo.GeoProxy(copy.deepcopy(GEOMETRY_COLLECTION))
    p.filter(validate)
    assert p.geotype == "GeometryCollection"
    assert p.root["geometries"] == [POINT]

    gc2 = copy.deepcopy(GEOMETRY_COLLECTION)
    gc2["geometries"] = [LINE_STRING, POLYGON_0]
    p = geo.GeoProxy(gc2)
    p.filter(validate)
    assert p.geotype is None


def test_filter_function_feature_collection():
    fc1 = copy.deepcopy(FEATURE_COLLECTION)
    point_feature = copy.deepcopy(FEATURE_1)
    point_feature["geometry"] = POINT
    fc1["features"].append(point_feature)
    p = geo.GeoProxy(fc1)
    p.filter(validate)
    assert p.geotype == "FeatureCollection"
    assert p.root["features"] == [point_feature]

    p = geo.GeoProxy(copy.deepcopy(FEATURE_COLLECTION))
    p.filter(validate)
    assert p.geotype is None


def test_polygon_from_hatch_hole_in_hole():
    hatch = factory.new("HATCH")
    paths = hatch.paths
    paths.add_polyline_path(square(10), flags=1)
    paths.add_polyline_path(translate(square(8), (1, 1)), flags=0)
    paths.add_polyline_path(translate(square(6), (2, 2)), flags=0)
    mapping = geo.proxy(hatch).__geo_interface__
    assert mapping["type"] == "Polygon"
    assert len(mapping["coordinates"]) == 2, "inner hole should be removed"

    mapping = geo.proxy(hatch, force_line_string=True).__geo_interface__
    assert mapping["type"] == "MultiLineString"
    assert len(mapping["coordinates"]) == 3, "inner hole should not be removed"


def test_three_polygons_from_one_hatch():
    hatch = factory.new("HATCH")
    paths = hatch.paths
    paths.add_polyline_path(square(1), flags=1)
    paths.add_polyline_path(translate(square(1), (3, 1)), flags=1)
    paths.add_polyline_path(translate(square(1), (6, 2)), flags=1)
    mapping = geo.proxy(hatch).__geo_interface__
    assert mapping["type"] == "MultiPolygon"
    assert len(mapping["coordinates"]) == 3


@pytest.mark.parametrize(
    "radius",
    [
        pytest.param(0.04999999, id="less-than-half-of-flattening-distance"),
        pytest.param(0.05, id="half-of-flattening-distance"),
        pytest.param(0.05000001, id="more-than-half-of-flattening-distance"),
    ],
)
def test_from_dxf_entities_given_arc_returns_linestring(radius: float) -> None:
    drawing = ezdxf.new()
    msp = drawing.modelspace()
    arc = msp.add_arc(
        center=(0, 0, 0),
        radius=radius,
        start_angle=0,
        end_angle=90,
    )
    proxy = geo.GeoProxy.from_dxf_entities(arc, distance=geo.MAX_FLATTENING_DISTANCE)
    assert proxy.__geo_interface__["type"] == "LineString"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import Dict

from ezdxf.document import Drawing
from ezdxf.tools import codepage

from ezdxf.sections.header import HeaderSection
from ezdxf.sections.classes import ClassesSection
from ezdxf.sections.tables import TablesSection
from ezdxf.sections.blocks import BlocksSection
from ezdxf.sections.entities import EntitySection
from ezdxf.sections.objects import ObjectsSection
from ezdxf.sections.acdsdata import AcDsDataSection

from .const import *
from .fileheader import FileHeader
from .header_section import load_header_section
from .classes_section import load_classes_section

__all__ = ["readfile", "load"]


def readfile(filename: str, crc_check=False) -> "Drawing":
    data = open(filename, "rb").read()
    return load(data, crc_check)


def load(data: bytes, crc_check=False) -> Drawing:
    doc = DwgDocument(data, crc_check=crc_check)
    doc.load()
    return doc.doc


class DwgDocument:
    def __init__(self, data: Bytes, crc_check=False):
        self.data = memoryview(data)
        self.crc_check = crc_check
        self.specs = FileHeader(data, crc_check=crc_check)
        self.doc: Drawing = self._setup_doc()
        # Store DXF object types by class number:
        self.dxf_object_types: Dict[int, str] = dict()

    def _setup_doc(self) -> Drawing:
        doc = Drawing(dxfversion=self.specs.version)
        doc.encoding = self.specs.encoding
        doc.header = HeaderSection.new()

        # Setup basic header variables not stored in the header section of the DWG file.
        doc.header["$ACADVER"] = self.specs.version
        doc.header["$ACADMAINTVER"] = self.specs.maintenance_release_version
        doc.header["$DWGCODEPAGE"] = codepage.tocodepage(self.specs.encoding)

        doc.classes = ClassesSection(doc)
        # doc.tables = TablesSection(doc)
        # doc.blocks = BlocksSection(doc)
        # doc.entities = EntitySection(doc)
        # doc.objects = ObjectsSection(doc)
        # doc.acdsdata = AcDsDataSection(doc)
        return doc

    def load(self):
        self.load_header()
        self.load_classes()
        self.load_objects()
        self.store_objects()

    def load_header(self) -> None:
        hdr_section = load_header_section(self.specs, self.data, self.crc_check)
        hdr_vars = hdr_section.load_header_vars()
        self.set_header_vars(hdr_vars)

    def set_header_vars(self, hdr_vars: Dict):
        pass

    def load_classes(self) -> None:
        cls_section = load_classes_section(
            self.specs, self.data, self.crc_check
        )
        for class_num, dxfclass in cls_section.load_classes():
            self.doc.classes.register(dxfclass)
            self.dxf_object_types[class_num] = dxfclass.dxf.name

    def load_objects(self) -> None:
        pass

    def store_objects(self) -> None:
        pass
